/**
/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2015 Voiceworks B.V. info@voiceworks.com
 *
 */

define('com.voiceworks/ox-messenger/conversation/main', [
    'com.voiceworks/ox-messenger/aggregator',
    'com.voiceworks/ox-messenger/core/session',
    'com.voiceworks/ox-messenger/util',
    'com.voiceworks/ox-messenger/history/main',
    'com.voiceworks/ox-messenger/conversation/models/conversation',
    'com.voiceworks/ox-messenger/conversation/views/conversation',
    'com.voiceworks/ox-messenger/conversation/views/conversationlist'
], function (Aggregator, Session, Util, History, ConversationModel, ConversationView,
        ConversationListView) {
    'use strict';

    /**
     * Conversation Controller
     */
    var ConversationController = {

        init: function (app) {
            this.app = app;

            Aggregator.on('connection:reconnecting', $.proxy(this.reopenConversation, this));
            ox.on('relogin:success', $.proxy(this.reopenConversation, this));
        },

        reopenConversation: function () {
            var self = this;

            // save the jid/muc id for later use
            var conversationId = Session.user.conversations.currentConversationId;

            // reset the currentConversationId to force re-opening of the conversation View
            Session.user.conversations.currentConversationId = null;

            // When MUC history is loaded
            Aggregator.on('muchistory:loaded', function () {
                var newConversation = Session.user.conversations.get(conversationId);
                if (newConversation && ox.ui.App.getCurrentApp().get('name') === 'com.voiceworks/ox-messenger') {
                    self.showConversation(newConversation);
                }
            });

            // When XMPP history is loaded
            Aggregator.on('xmpp-history:loaded', function () {
                var newConversation = Session.user.conversations.get(conversationId);
                if (newConversation && ox.ui.App.getCurrentApp().get('name') === 'com.voiceworks/ox-messenger') {
                    self.showConversation(newConversation);
                }
            });
        },

        initConversation: function ($conversationEl) {
            this.$conversationEl = $conversationEl;
        },

        initList: function ($listEl) {
            this.$listEl = $listEl;
        },

        showList: function () {
            var that = this;
            this.listView && this.listView.remove() && delete this.listView;
            this.listView = new ConversationListView({
                app: this.app,
                ignoreFocus: true,
                pagination: false,
                draggable: false
            });
            this.listView.setCollection(Session.user.conversations);
            this.$listEl.append(this.listView.render().el);
            this.listView.$el.off('mousedown.dnd');
            this.listView.on('selection:change', function (ids) {
                if (ids.length === 1){
                    that.showConversation(Session.user.conversations.get(ids[0]), true);
                }
                Aggregator.trigger('conversationlist:selection:change', ids);
            });
            return this;
        },

        conversationDeleted: function () {
            if (this.conversationView) {
                this.conversationView.remove() && delete this.conversationView;
            }
        },

        showConversation: function (conversation, dontTrigger, forceShow) {
            var self = this;

            History.hide();

            if (!conversation) {
                return;
            }

            if (Session.user.conversations.currentConversationId === conversation.id &&
                !forceShow) {
                conversation.set('scroll', true);
                return;
            }
            Session.user.conversations.currentConversationId = conversation.id;

            // Save streams temp
            this.keepCurrentAlive();

            if (this.conversationView) {
                this.conversationView.remove();
                this.conversationView = null;
            }

            this.conversationView = new ConversationView({
                user: Session.user,
                app: this.app,
                model: conversation
            });

            // Container for showing the conversation
            var windowRightContainer = this.$conversationEl.find('.messenger--window-right-container');
            if (windowRightContainer.length > 0) {
                windowRightContainer.children().remove();
                windowRightContainer.prevObject = null;
                windowRightContainer.append(this.conversationView.render().$el);
            } else {
                windowRightContainer = $('<div class="messenger--window-right-container">');
                windowRightContainer.append(this.conversationView.render().$el);
            }
            this.$conversationEl.append(windowRightContainer);

            // Compact View? Add correct class
            if (this.app.props.get('layout') === 'compact') {
                windowRightContainer.addClass('messenger--window-right-container--compact');
            }

            var conversationListIndex = Session.user.conversations.indexOf(conversation);
            if (!dontTrigger && this.listView && this.listView.selection.get() !== conversationListIndex) {
                //this.listView.selection.select(conversationListIndex);
            }

            conversation.setUnreadMessages(0);
            this.conversationView.messageFormView.focusInput(true);
            this.conversationView.messageContainerCollectionView.scroll(true);

            if ($('div.messenger--chat-messages').scrollTop() === 0 && conversation.messageArchives.length) {
                this.preloadArchives(conversation);
            }

            $('div.messenger--chat-messages').scroll(function () {
                if ($(this).scrollTop() === 0 && conversation.messageArchives.length) {
                    $(this).scrollTop(20);
                    conversation.getNextArchiveMessage();
                }
            });

            $(document).ready(function () {
                if (self.conversationView) {
                    self.conversationView.setMessagingHeight();
                }
                $(window).on('resize', function () {
                    if (self.conversationView) {
                        self.conversationView.setMessagingHeight();
                    }
                });
            });
        },

        // TODO move the UI logic to the view
        preloadArchives: function (conversation) {
            var that = this;
            conversation.getNextArchiveMessage().then(function () {
                // Check if the conversation is still opened.
                var container = $('div.messenger--chat-messages');
                if (!container.length) {
                    return;
                }
                container.scrollTop(container[0].scrollHeight + 10);

                if (container.scrollTop() < 1 && conversation.messageArchives.length) {
                    that.preloadArchives(conversation);
                }
            });
        },

        delete: function (conversationId) {
            Session.user.conversations.remove(Session.user.conversations.get(conversationId));
            Session.user.conversations.trigger('reset');
            this.conversationDeleted();
        },

        start: function (participant, type) {
            var conversation = Session.user.conversations.get(participant.id);

            if (conversation) {
                return conversation;
            }

            conversation = new ConversationModel({
                id: participant.id,
                title: participant.getDisplayName(),
                user: Session.user,
                sortName: participant.getDisplayName(),
                timestamp: new Date().getTime()
            });

            conversation.addParticipant(participant);
            conversation.addParticipant(Session.user);
            Session.user.conversations.add(conversation);

            return conversation;
        },

        // it keeps the current conversation's streams alive
        keepCurrentAlive: function () {

            var activeCallSession = Session.user.call.get('callSession');

            // Save steams temp
            var streamLocal = $('#videoLocal'),
                streamRemote = $('#videoRemote'),
                groupStreamVideos = $('.groupStreamVideo');

            if (activeCallSession) {
                var activeCallConversation = Session.user.getCurrentCallConversation();

                if (!$('#videoTmp').length) {
                    $('body').append(
                        $('<div id="videoTmp">')
                    );
                }

                if (activeCallConversation.isMuc()) {
                    $('#videoTmp').children().remove();
                    $('#videoTmp').append(
                        streamLocal,
                        groupStreamVideos
                    );

                    // If it's not a video element, it's a flash object, so there's no
                    // @TODO-FLASH what's happening here?
                    if (streamLocal.length && streamLocal.is('video')) {
                        streamLocal.get(0).play();
                    }

                    $.each(groupStreamVideos, function (index, video) {
                        video.play();
                    });

                } else {
                    $('#videoTmp').children().remove();
                    $('#videoTmp').append(
                        streamLocal,
                        streamRemote
                    );

                    // If it's not a video element, it's a flash object, so there's no 'play' method
                    // @TODO-FLASH what's happening here?
                    if (streamLocal.length && streamLocal.is('video')) {
                        streamLocal.get(0).play();
                    }
                    if (streamRemote.length && streamRemote.is('video')) {
                        streamRemote.get(0).play();
                    }
                }
            }
        }
    };

    return ConversationController;
});
