/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2015 Voiceworks B.V. info@voiceworks.com
 *
 */

define('com.voiceworks/ox-messenger/conversation/models/conversation', [
    'com.voiceworks/ox-messenger/aggregator',
    'com.voiceworks/ox-messenger/roster/models/usercollection',
    'com.voiceworks/ox-messenger/conversation/models/messagecontainercollection',
    'com.voiceworks/ox-messenger/conversation/models/messagecontainer',
    'com.voiceworks/ox-messenger/conversation/models/message/factory',
    'com.voiceworks/ox-messenger/conversation/models/message/system',
    'com.voiceworks/ox-messenger/conversation/models/messagearchivecollection',
    'com.voiceworks/ox-messenger/conversation/models/messagearchive',
    'com.voiceworks/ox-messenger/util',
    'com.voiceworks/ox-messenger/core/notification-control',
    'com.voiceworks/ox-messenger/core/notification',
    'com.voiceworks/ox-messenger/core/properties',
    'com.voiceworks/ox-messenger/core/vendor/promise',
    'com.voiceworks/ox-messenger/conversation/models/message',
    'settings!com.voiceworks/ox-messenger',
    'gettext!com.voiceworks/ox-messenger'
], function (Aggregator, RosterUserCollection, MessageContainerCollection, MessageContainerModel,
             MessageFactory, SystemMessageModel, MessageArchiveCollection, MessageArchiveModel, Util,
             NotificationControl, Notification, PropertiesModel, Promise, MessageModel, settings, gt) {

    'use strict';

    /**
     *  Conversation
     *    Message Containers Collection
     *      Message Container
     *        Messages Collection
     *          Message
     *    Message Archive Collection
     */


    /**
     * Conversation Model
     */
    var ConversationModel = Backbone.Model.extend({

        defaults: {
            id: '',
            title: 'default chat',
            timestamp: 0,
            lastMsg: new MessageModel({message: '...', timestamp: -1}),
            previousMsgJid: false,
            callSession: null,
            callState: 'idle',
            unreadMessages: 0,
            scroll: false,
            timer: null,
            sortName: '',
            showNotifications: true,
            chatState: '',                  // contains the chat state ('paused', 'active', 'composing',..) for logged user
            composingTimeout: null,         // timeout of the logged user to manage the 'composing'/'paused' state
            composingParticipants: [],      // array with the participants of the chat that are typing (composing)
            whosComposingText: '',          // list of names of the users typing (composing)

            /**
             * Colleciton of archive collections, represented by the full jid and the start date.
             * @type {Array.<{jid: string, date: Date}>}
             */
            jidToDate: null
        },

        initialize: function (options) {
            this.user = options.user;
            this.participants = new RosterUserCollection();
            this.participants.sortFirst = this.user;
            this.callParticipants = new RosterUserCollection();
            this.messageContainers = this.createMessagesCollection();
            this.messagesCollection = new Backbone.Collection();
            this.messagesCollection.comparator = 'timestamp';
            this.messageArchives = new MessageArchiveCollection();
            this.groupCallInvitees = new RosterUserCollection();
            this.chatState = '';
            this.composingTimeout = null;
            this.composingParticipants = [];
            this.whosComposingText = '';

            this.set('jidToDate', []);
            this.set('showNotifications', true);

            this.properties = new PropertiesModel({
                groupMember: this.isMember(this.user),
                showNotifications: this.get('showNotifications')
            });

            this.participantLimit = {
                video: 5,
                audio: 20
            };

            Aggregator.on('conversation:showNotifications', $.proxy(this.handleShowNotifications, this));
        },

        createMessagesCollection: function () {
            var collection = new MessageContainerCollection();
            var conversationModel = this;
            // TODO: this is a hack. FIXME!
            Object.defineProperty(collection, 'models', {
                get: function () {
                    // Dynamically create group containers from a flat msgs list
                    return conversationModel.messagesCollection.reduce(function (ret, item, index) {
                        // Create a new group container, when required
                        if (!ret.length ||
                            ret[ret.length - 1].get('senderUser') !== item.user ||
                            ret[ret.length - 1].has('fileDesc') ||
                            item.has('fileDesc')) { // a group container per file

                            ret.push(new MessageContainerModel({
                                senderUser: item.user,
                                fileDesc: item.get('fileDesc'),
                                // TODO: WTF?
                                conversation: conversationModel
                            }));
                        }

                        // Add the message to the last container
                        ret[ret.length - 1].messages.add(item);

                        // Repeat
                        return ret;
                    }, []).map(function (item) {
                        item.messages.sort();
                        return item;
                    });
                }
            });

            return collection;
        },

        /**
         * @param message
         * @TODO rename addMessage => handleMessage or something else?
         */
        addMessage: function (message) {
            if (message.get('isOfflineMessage') || !message.get('isHistory') && !MessageFactory.isSystem(message)) {
                this.handleMessageNotificiations(this, message);
            }

            this.messagesCollection.add(message);
            this.messagesCollection.sort();

            // Update the last message if it's not a System Message
            if(this.messagesCollection.length > 0){
                var lastMessage = this.messagesCollection.at(this.messagesCollection.length - 1);

                // TODO this.lastMessageHandler();
                if (!MessageFactory.isSystem(lastMessage)) {
                    this.set('lastMsg', lastMessage);
                    this.set('timestamp', lastMessage.get('timestamp'));
                }
            }

            this.messageContainers.trigger('add');
        },

        getNextArchiveMessage: function () {
            var archive = this.messageArchives.pop(),
                addMessage = this.addMessage.bind(this);

            return new Promise(function (resolve, reject) {
                if (archive) {
                    archive.getMessageModels(function (msgs) {
                        msgs.forEach(function (item) {
                            // Prepend
                            addMessage(item);
                        });
                        resolve();
                    });
                } else {
                    resolve();
                }
            });
        },

        addParticipant: function (participant, isHistory, avoidSystemMessage) {
            // Make sure the participant does not exist already in the conversation
            if (!this.existsBareParticipant(participant.id)) {
                this.participants.add(participant);

                if(this.isMuc() && !avoidSystemMessage){
                    var joinMessage = participant.id === this.user.id ? gt('You joined the group conversation') : gt('%s joined the group conversation', participant.getDisplayName()),
                        sysMessage = new SystemMessageModel({
                            message: joinMessage,
                            timestamp: new Date().getTime()
                        });

                    this.addMessage(sysMessage);
                }
            }
        },

        removeParticipant: function (participant) {
            var isLoggedUser = participant.id === this.user.id;

            this.participants.remove(participant);

            // Clear the conversation when I left it
            if(isLoggedUser)
                this.messagesCollection.reset();

            // If the participant was properly deleted, create a system message
            if (!this.existsBareParticipant(participant.id)) {
                var sysMessage = new SystemMessageModel({
                    message: isLoggedUser?gt('You left the group conversation') : gt('%s left the group conversation', participant.getDisplayName()),
                    timestamp: new Date().getTime()
                });

                this.addMessage(sysMessage);
            }
        },

        getOtherParticipants: function () {
            var cloned;

            if (!this.user) {
                return this.participants;
            }

            // return without initiator
            cloned = this.participants.clone();
            cloned.models = _(this.participants.models).without(this.user);

            return cloned;
        },

        getParticipantsCount: function () {
            if (!this.participants) {
                return;
            }

            return this.participants.length;
        },

        getCallParticipantsCount: function () {
            if (!this.callParticipants) {
                return;
            }

            return this.callParticipants.length;
        },

        getParticipantsOnlineCount: function () {
            var onlineParticipants = this.participants.filter(function (p) {
                return p.isOnline() === true;
            });

            return onlineParticipants.length;
        },

        getParticipantsActiveCallCount: function () {
            var callParticipants = this.getCallParticipants();

            return callParticipants.length;
        },

        existsBareParticipant: function (jid) {
            var exists = false;

            this.participants.each(function (participant) {
                exists = exists || (Util.getBareJid(participant.id) === Util.getBareJid(jid));
            });

            return exists;
        },

        handleMessageNotificiations: function (conversation, message) {
            var unreadMessages = conversation.get('unreadMessages'),
                previousCount = unreadMessages ? unreadMessages : 0,
                newCount = 0,
                appActive = ox.ui.App.getCurrentApp().get('name') === 'com.voiceworks/ox-messenger';

            // Reset the count (= 0) when the conversation is the current open conversation
            if (appActive && conversation.user.conversations.currentConversationId === conversation.get('id')) {
                newCount = 0;

                // Don't add number if it's your own message
            } else if (appActive && conversation.user.get('id') === message.getSenderJid()) {
                newCount = unreadMessages;

                // Not current conversation and not your own message? Add 1 to the unreadMessages count
            } else {
                newCount = previousCount + 1;

                // Play sound for incoming message, unless the user is DND or if notifications are muted.
                NotificationControl.notify(conversation, {
                    sound: function () {
                        Util.playSound(settings.get('incomingMessageTone'), 'incomingMessagePing', false);
                    },
                    yell: function () {
                        Notification.yell(message.user, message);
                    }
                });
            }

            this.setUnreadMessages(newCount);
            this.collection.updateAppUnreadBadge();
        },

        setUnreadMessages: function (newCount) {
            this.set('unreadMessages', newCount);
            this.collection.updateAppUnreadBadge();
        },

        isMuc: function () {
            return this.get('roomId') ? true : false;
        },

        createCallHistory: function (callHistory) {
            var that = this,
                messages = [];

            callHistory.each(function (cdr) {
                messages.push(that.createCallHistoryItem(cdr));
            });

            return messages;
        },

        createCallHistoryItem: function (cdr) {
            var message,
                duration = Util.getDurationFormatted(cdr.get('billmsec') / 1000);

            switch (cdr.getCallHangupCause()) {
                case 'cancelled':
                    message = gt('Cancelled call');
                    break;
                case 'missed':
                    message = gt('Missed call');
                    break;
                default:
                    message = gt('Call - duration: %s', duration);
            }

            message = new SystemMessageModel({
                from: 'callhistory',
                message: message,
                timestamp: Date.parse(cdr.getStartStamp()),
                isHistory: true
            });

            return message;
        },

        hasActiveCall: function () {
            return this.getParticipantsActiveCallCount() > 1;
        },

        isMember: function (user) {
            return this.participants.get(user) ? true : false;
        },

        canGroupCall: function () {
            var that = this,
                errorMessages = {
                    video: [],
                    audio: [],
                    generic: []
                },
                isMember = this.isMember(this.user),
                callableParticipants = 0;

            /**
             * Generic Checks
             */
            if (!isMember) {
                errorMessages.generic.push(gt('You are not a member of the group chat.'));
            }

            if (this.getParticipantsCount() < 2) {
                errorMessages.generic.push(gt('You are the only member in the group chat.'));
            }

            if (this.user.call.isActive()) {
                errorMessages.generic.push(gt('You are already in a call.'));
            }

            if (this.user.get('mustFlash')) {
                errorMessages.generic.push('Calling is only supported using Chrome or Firefox.');
            }

            /**
             * Group Call Checks
             */
            var groupCallActive = false;
            var callType = this.get('callType') || 'audio';
            this.participants.each(function (participant) {
                if (participant === that.user) {
                    return;
                }

                var canCallWith = that.user.canCallWith(callType, participant, true);

                if (canCallWith === true) {
                    callableParticipants++;
                }

                if (participant.get('inGroupCall') && participant.get('inGroupCall') === that.get('roomId')) {
                    groupCallActive = true;
                }
            });

            /**
             * Checks when there is an active call in this conversation
             */
            if (this.hasActiveCall()) {

                var callParticipants = this.getCallParticipants();

                // Disable video buttons when maximum amount of users has been met.
                if (callParticipants.length >= this.participantLimit.video) {
                    errorMessages.video.push(gt('The maximum amount of %d active participants is reached.', this.participantLimit.video));
                }

                // Disable audio buttons when maximum amount of users has been met.
                if (callParticipants.length >= this.participantLimit.audio) {
                    errorMessages.audio.push(gt('The maximum amount of %d active participants is reached.', this.participantLimit.audio));
                }

                /**
                 * Checks when there is NOT an active call in this conversation
                 */
            } else {
                // Disable video buttons when maximum amount of users has been met.
                if (this.getParticipantsOnlineCount() > this.participantLimit.video) {
                    errorMessages.video.push(gt('You can not start a group video call with more than %d participants.', this.participantLimit.video));
                }

                // Disable audio buttons when maximum amount of users has been met.
                if (this.getParticipantsOnlineCount() > this.participantLimit.audio) {
                    errorMessages.audio.push(gt('You can not start a group audio call with more than %d participants.', this.participantLimit.audio));
                }

            }

            // Disable buttons when none of the participants is able to call (busy)
            if (callableParticipants === 0 && !groupCallActive && !this.user.get('mustFlash')) {
                errorMessages.generic.push(gt.ngettext(
                    'Group call not possible: a participant is busy or not able to call.',
                    'Group call not possible: all participants are busy or not able to call.',
                    this.participants - 1
                ));
            }

            errorMessages.audio = errorMessages.audio.concat(errorMessages.generic);
            errorMessages.video = errorMessages.video.concat(errorMessages.generic);

            if (errorMessages.audio.length || errorMessages.video.length) {
                return errorMessages;
            }

            return true;
        },

        canJoinGroupCall: function (callType) {
            var maxCallParticipants = (callType === 'groupvideo') ?  this.participantLimit.video : this.participantLimit.audio;
            var callParticipants = this.getCallParticipants();
            return callParticipants.length < maxCallParticipants;
        },

        sendGroupCallInvite: function (user) {
            if (!this.groupCallInvitees.get(user)) {
                return false;
            }

            Aggregator.trigger('groupcall:sendInvite', {
                conversation: this,
                user: user,
                notify: true
            });

            this.groupCallInvitees.remove(user);

            return true;
        },

        // Local setting for conversation to mute notifications
        isMuted: function () {
            return !this.get('showNotifications');
        },

        handleShowNotifications: function (object) {
            if (this.cid !== object.conversation.cid) {
                return;
            }

            this.set('showNotifications', object.showChatNotifications);
        },

        // Returns the number of participants that are 'ready' on the conversation
        getNumParticipantsReady: function () {
            return this.participants
                .filter(function (participant) {return participant.call && participant.call.get('readyReceived'); })
                .length;
        },

        // Returns the number of participants that are 'ready' on the conversation
        getNumParticipantsInGroupCall: function () {
            var thisRoomId = this.get('roomId');

            return this.participants
                .filter(function (participant) {
                    return (participant.get('inGroupCall') && participant.get('inGroupCall') === thisRoomId);
                }).length;
        },

        // Starts a activity timeout that will callback whether there's activity (anybody 'ready') or not in this
        // conversation right after 60s
        startActivityTimeout: function (cb) {
            var conversation = this,
                timeoutTime = 60000;

            setTimeout(function () {
                cb(conversation.getNumParticipantsReady() > 1);
            }, timeoutTime);
        },

        addComposingParticipant: function(user){
            // Make sure we don't duplicate the participant
            this.removeComposingParticipant(user);

            // Add participant
            this.composingParticipants.push(user);

            this.updateWhoisComposing();
        },

        removeComposingParticipant: function(user){
            // Remove participant
            if(this.composingParticipants.indexOf(user) >= 0){
                this.composingParticipants.splice(this.composingParticipants.indexOf(user), 1);
            }

            this.updateWhoisComposing();
        },

        setChatState: function (state) {
            this.chatState = state;

            if(state === ConversationModel.STATE_ACTIVE){
                if(this.clearChatStateTimeout()){
                    return;
                }
            } else if(state === ConversationModel.STATE_PAUSED){
                this.clearChatStateTimeout();
            } else if(state === ConversationModel.STATE_COMPOSING){
                // If the timeout was not in progress, we'll trigger the event
                if(this.resetComposingTimeout()){
                    return;
                }
            }

            // Trigger the state change
            Aggregator.trigger('conversation:state', {
                conversation: this,
                state: state
            });
        },

        // Provided a fileDesc, it returns the first and (hopefully) only message with that file desc
        getMessageByFileDesc: function(fileDesc) {
            var theMessage = _.find(this.messagesCollection.models, function(message){
                return (message.has('fileDesc') && message.get('fileDesc').cid === fileDesc.cid);
            });

            return theMessage;
        },

        // Resets the timer to trigger 'conversation:state' paused message
        resetComposingTimeout: function() {
            var timeoutCleared,
                onTimeout = function(){
                    // Dispatch a 'paused' event
                    this.setChatState(ConversationModel.STATE_PAUSED);
                    clearTimeout(this.composingTimeout);
                };

            timeoutCleared = this.clearChatStateTimeout();
            this.composingTimeout = setTimeout($.proxy(onTimeout, this), ConversationModel.STATE_PAUSED_TIME);

            return timeoutCleared;
        },

        clearChatStateTimeout: function () {
            if(this.composingTimeout){
                clearTimeout(this.composingTimeout);
                this.composingTimeout = null;
                return true;
            }
            return false;
        },

        /**
         * TODO move lang strings to the lang file
         */
        updateWhoisComposing: function() {
            var text = '';

            if(this.composingParticipants.length === 1){
                text = gt('%s is typing...',
                    this.composingParticipants[0].getDisplayName());
            }
            else if(this.composingParticipants.length === 2){
                text = gt('%s and %s are typing...',
                    this.composingParticipants[0].getDisplayName(),
                    this.composingParticipants[1].getDisplayName());
            }
            else if(this.composingParticipants.length > 0){
                // TODO show the number of others
                text = gt('%s, %s and # others are typing...',
                    this.composingParticipants[0].getDisplayName(),
                    this.composingParticipants[1].getDisplayName());
            }

            this.set('whosComposingText', text);
        }
    });

    // ms to consider the chat must be switched to PAUSED
    ConversationModel.STATE_PAUSED_TIME = 15000;

    // Chat states (per user)
    ConversationModel.STATE_PAUSED      = 'paused';
    ConversationModel.STATE_ACTIVE      = 'active';
    ConversationModel.STATE_COMPOSING   = 'composing';

    return ConversationModel;
});
