/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2015 Voiceworks B.V. info@voiceworks.com
 *
 */

define('com.voiceworks/ox-messenger/conversation/views/stream', [
    'com.voiceworks/ox-messenger/core/views/disposable',
    'com.voiceworks/ox-messenger/core/views/partials/callpart',
    'com.voiceworks/ox-messenger/core/views/timer'
], function (DisposableView, CallPart, TimerView) {
    'use strict';

    var StreamView = DisposableView.extend({
        className: 'messenger--table__cell  messenger--table__cell--centered  messenger--call-group-video__participant',

        initialize: function (options) {
            DisposableView.prototype.initialize.apply(this, arguments);

            this.id = this.model.id;
            this.title = options.title || this.model.getDisplayName();
            this.listenTo(this.model.call, 'change:callState', this.stateCheck);

            this.listenTo(this.model, 'change:streamFrozen', this.handleFrozenUnFrozenStream);
        },

        render: function () {
            var infoDiv = $('<div class="messenger--call-info">');

            this.$el.children().remove();
            this.$el.prevObject = null;

            this.$el.attr('id', 'stream-' + this.model.id.replace(/[@|.]/g, '_'));

            if (this.model.call.get('type') === 'groupvideo') {
                // Add group-video specific Elements
                this.$el.append(
                    $('<div class="messenger--call-group-video__video-container  messenger--is-hidden">').append(
                        $('<div class="messenger--call-group-video__frozen  messenger--is-hidden">').append(
                            $('<i class="fa fa-spinner fa-spin">')
                        ),
                        $('<div class="messenger--call-group-video__video">')
                    ),
                    $('<div class="messenger--call-group-video__video-name  messenger--text-wrap  messenger--is-hidden">').text(
                        this.title
                    )
                );

                // Setup info Element
                infoDiv.addClass('messenger--call-group-video__call-info');
                // Stream avatar
                infoDiv.append(
                    $('<div class="messenger--avatar  messenger--avatar--extra-large  messenger--call-group-video__avatar">').css(
                        {'background-image': 'url(' + this.model.getImageUrl() + ')'}
                    )
                );
            } else {
                // Setup info Element
                infoDiv.addClass('messenger--call-group-audio__call-info');
                // Stream avatar
                infoDiv.append(
                    $('<div class="messenger--avatar  messenger--avatar--extra-large  messenger--call-group-video__avatar  messenger--avatar--icon  messenger--avatar--icon--large">').append(
                        $('<div class="fa fa-group"></div>')
                    )
                );
            }

            // Finish info Element
            infoDiv.append(
                $('<div class="messenger--call-group__name  messenger--text-wrap  messenger--font--bold">').text(this.title),
                $('<div class="messenger--call-group__status  messenger--font--light">'),
                CallPart.getConnectingAnimation()
            );

            // Add info element to View
            this.$el.append(infoDiv);

            // Update View
            this.stateCheck();

            return this;
        },

        stateCheck: function () {
            var callState = this.model.call.get('callState');

            switch (callState) {
                case 'idle':
                case 'calling':
                    this.stateIsIdle();
                    break;
                case 'in-progress':
                    this.stateIsInProgress();
                    break;
                case 'group-initiated':
                    this.stateIsGroupInitiated();
                    break;
                case 'answered':
                    this.stateIsAnswered();
                    break;
            }
        },

        // idle: hangup
        // hide all
        // show nothing
        stateIsIdle: function () {
            this.$el.find('.messenger--call-group-video__video-container video').remove();
            this.showVideoContainer(false);

            // this is for calling:
            this.showInfo(true);
            this.setStatusLabel(this.model.call.getCallStateDisplayName());
            this.model.set('lastCallFrozen', this.model.get('streamFrozen'));
            this.model.set('streamFrozen', false);
        },

        // in-progress: not yet answered
        // Show Ringing status
        stateIsInProgress: function () {
            this.showVideoContainer(false);
            this.showInfo(true);
            this.setStatusLabel(this.model.call.getCallStateDisplayName());
        },

        // group-initiated: Only when it's my stream on a group audio call
        stateIsGroupInitiated: function () {
            // Only processed if it's an group audio call
            if (this.model.call.get('type') === 'groupaudio') {
                var timerView = new TimerView({
                        model: this.model.call.timer
                    });

                // Update the status label while waiting for media
                this.setStatusLabel(this.model.call.getCallStateDisplayName());

                // TODO: Handle 'canplay' event as we do with video calls
                // Firefox & Chrome handle that event in a different way and the change must be done on UCConnect level
                // Hide call connecting animation
                this.showCallConnecting(false);

                // Add the timer as status
                this.$el.find('.messenger--call-group__status')
                    .empty()
                    .append(timerView.render().$el);
            }
        },

        // answered: call started
        // show stream and add video to stream, and show timer
        // hide icon and status
        stateIsAnswered: function () {
            this.media = this.model.call.get('callSession').getMediaCanvas();
            var media = this.media,
                videoElem = this.$el.find('.messenger--call-group-video__video'),
                streamId = 'video-stream-' + this.model.id.replace(/[@|.]/g, '_'),
                tmpVideoStream = $('#' + streamId),
                onCanPlay = function () {
                    // When Group Video Call
                    if (this.model.call.get('type') === 'groupvideo') {
                        this.showVideoContainer(true);
                        this.showVideoName(true);
                        this.showInfo(false);

                        // When Group Audio Call
                    } else if (this.model.call.get('type') === 'groupaudio') {
                        this.setStatusLabel(this.model.call.getCallStateDisplayName());
                        this.showInfo(true);
                        this.showCallConnecting(false);
                    }
                    // Listen to the 'canplay' to wait for media
                    media.remote.removeEventListener('canplay', onCanPlay);
                };

            // There is a video saved, and it's a MUC
            if (tmpVideoStream.length) {
                media.remote = tmpVideoStream.get(0);
                tmpVideoStream.remove();
            }
            media.remote.setAttribute('id', streamId);
            media.remote.setAttribute('class', 'groupStreamVideo');

            if (media.remote) {
                this.progressListener = this.progressHandler.bind(this);
                this.timeupdateListener = this.timeupdateHandler.bind(this);
                $(media.remote)
                    .on('progress', this.progressListener)
                    .on('timeupdate', this.timeupdateListener);
            }

            videoElem.empty().append(media.remote);

            // Listen to the 'canplay' to wait for media
            media.remote.addEventListener('canplay', $.proxy(onCanPlay, this), false);

            if (media.remote.paused) {
                this.setStatusLabel(this.model.call.getCallStateDisplayName());
                setTimeout(function () {
                    media.remote.play();
                }, 100);
            }
        },

        showVideoContainer: function (show) {
            if (show) {
                this.$el.find('.messenger--call-group-video__video-container').removeClass('messenger--is-hidden');
            }
            else {
                this.$el.find('.messenger--call-group-video__video-container').addClass('messenger--is-hidden');
            }
        },

        showVideoName: function (show) {
            if (show) {
                this.$el.find('.messenger--call-group-video__video-name').removeClass('messenger--is-hidden');
            }
            else {
                this.$el.find('.messenger--call-group-video__video-name').addClass('messenger--is-hidden');
            }
        },

        showInfo: function (show) {
            if (show) {
                this.$el.find('.messenger--call-info').removeClass('messenger--is-hidden');
            }
            else {
                this.$el.find('.messenger--call-info').addClass('messenger--is-hidden');
            }
        },

        showCallConnecting: function (show) {
            if (show) {
                this.$el.find('.messenger--throbber.call--connecting').removeClass('messenger--is-hidden');
            }
            else {
                this.$el.find('.messenger--throbber.call--connecting').addClass('messenger--is-hidden');
            }
        },

        setStatusLabel: function (txt) {
            this.$el.find('.messenger--call-group__status').text(txt);
        },

        timeupdateHandler: function (event) {
            this.model.set('streamCurrentTime', new Date().getTime());
        },

        progressHandler: function (event) {
            var tmpTime        = new Date().getTime(),
                tmpFrozenLimit = 2000,
                frozen         = this.model.get('streamFrozen'),
                currentTime    = this.model.get('streamCurrentTime');

            if ((tmpTime - currentTime) > tmpFrozenLimit && !frozen && currentTime !== 0) {
                this.model.set('streamFrozen', true);
            }
            else if ((tmpTime - currentTime) <= tmpFrozenLimit && frozen && currentTime !== 0) {
                this.model.set('streamFrozen', false);
            }
        },

        handleFrozenUnFrozenStream: function (event) {
            var frozen = this.model.get('streamFrozen');

            if (frozen) {
                this.$el.find('.messenger--call-group-video__frozen').removeClass('messenger--is-hidden');
            } else {
                this.$el.find('.messenger--call-group-video__frozen').addClass('messenger--is-hidden');
            }
        },

        removeProgressListener: function () {
            if (this.media) {
                $(this.media.remote)
                    .off('progress', this.progressListener)
                    .off('timeupdate', this.timeupdateListener);
                delete this.media;
            }
        },

        dispose: function() {
            // the super call
            DisposableView.prototype.dispose.call(this);
            this.removeProgressListener();
        }
    });

    return StreamView;
});
