/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2015 Voiceworks B.V. info@voiceworks.com
 *
 */

/* global Strophe */

define('com.voiceworks/ox-messenger/mediator', [
    'io.ox/contacts/api',
    'io.ox/core/notifications',
    'io.ox/core/capabilities',
    'io.ox/core/folder/api',
    'com.voiceworks/ox-messenger/aggregator',
    'com.voiceworks/ox-messenger/ucconnect',
    'com.voiceworks/ox-messenger/core/session',
    'com.voiceworks/ox-messenger/core/notification-control',
    'com.voiceworks/ox-messenger/core/notification',
    'com.voiceworks/ox-messenger/roster/main',
    'com.voiceworks/ox-messenger/phone/main',
    'com.voiceworks/ox-messenger/conversation/main',
    'com.voiceworks/ox-messenger/conversation/models/conversation',
    'com.voiceworks/ox-messenger/conversation/models/groupconversation',
    'com.voiceworks/ox-messenger/conversation/models/message/factory',
    'com.voiceworks/ox-messenger/conversation/models/messagearchive',
    'com.voiceworks/ox-messenger/util',
    'com.voiceworks/ox-messenger/desktop_notifications/main',
    'com.voiceworks/ox-messenger/history/main',
    'com.voiceworks/ox-messenger/history/models/history',
    'com.voiceworks/ox-messenger/core/views/partials/userpart',
    'gettext!com.voiceworks/ox-messenger',
    'settings!com.voiceworks/ox-messenger',
    'settings!io.ox/core'
], function (ContactsApi, CoreNotification, Capabilities, folderApi, Aggregator, UCConnect,
        Session, NotificationControl, Notification, Roster, Phone, Conversation, ConversationModel,
        GroupConversationModel, MessageFactory, MessageArchiveModel, Util, DesktopNotifications,
        CallHistory, HistoryModel, UserPart, gt, Settings, CoreSettings) {
    'use strict';

    /**
     * Enum with allowed call types.
     *
     * TODO move to a separate namespace.
     */
    var CALL_TYPE = {
        video: true,
        audio: true,
        groupvideo: true,
        groupaudio: true
    };

    var _currentAppName;

    function assert(condition, msg) {
        if (!condition) {
            throw new Error(msg || '');
        }
    }

    var Mediator = {
        init: function () {

            // TEMPORARY
            if (Settings.get('callHistoryLimit') === 1000) {
                Settings.set('callHistoryLimit', 200);
            }

            this.desktopNotification = DesktopNotifications.init();
            this.msgArchives = [];
            this.msgArchiveCount = [];

            var initialPresence = Settings.get('initialPresenceStatus');
            if (!initialPresence) {
                initialPresence = 'online';
            }

            UCConnect.init({
                initialPresence: initialPresence,
                logEnabled: _.url.hash('debug-messenger')
                //logConnector: $.proxy(Session.logger.write, this)
            });

            UCConnect.events.on('call_state', $.proxy(this.callStateHandler, this));
            UCConnect.events.on('connecting', $.proxy(this.connectingHandler, this));
            UCConnect.events.on('connected', $.proxy(this.connectedHandler, this));
            UCConnect.events.on('disconnected', $.proxy(this.disconnectedHandler, this));
            UCConnect.events.on('end_call', $.proxy(this.endCallHandler, this));
            UCConnect.events.on('failed', $.proxy(this.failedHandler, this));
            UCConnect.events.on('invite', $.proxy(this.groupInvitedHandler, this));
            UCConnect.events.on('local_media_requested', $.proxy(this.localMediaRequestedHandler, this));
            UCConnect.events.on('local_media_failed', $.proxy(this.localMediaFailedHandler, this));
            UCConnect.events.on('local_media_acquired', $.proxy(this.localMediaAcquiredHandler, this));
            UCConnect.events.on('message', $.proxy(this.messageHandler, this));
            UCConnect.events.on('memberlist', $.proxy(this.memberListHandler, this));
            UCConnect.events.on('mystatus', $.proxy(this.myStatusHandler, this));
            UCConnect.events.on('new_call', $.proxy(this.newCallHandler, this));
            UCConnect.events.on('parameters', $.proxy(this.parametersHandler, this));
            UCConnect.events.once('roster', $.proxy(this.initialRosterHandler, this));
            UCConnect.events.on('status', $.proxy(this.statusHandler, this));
            UCConnect.events.on('vcard_updated', $.proxy(this.vcardUpdatedHandler, this));
            UCConnect.events.on('chat_state', $.proxy(this.chatStateHandler, this));
            UCConnect.events.on('subscribe', $.proxy(this.subscribeHandler, this));
            UCConnect.events.on('archivecollection', $.proxy(this.archiveCollectionHandler, this));
            UCConnect.events.on('conversationdeleted', $.proxy(this.conversationDeletedHandler, this));
            UCConnect.events.on('groupcallinvite', $.proxy(this.groupCallInviteHandler, this));
            UCConnect.events.on('groupcallready', $.proxy(this.groupCallReadyHandler, this));
            UCConnect.events.on('groupcallcancelled', $.proxy(this.groupCallCancelledHandler, this));
            UCConnect.events.on('groupcalldeclined', $.proxy(this.groupCallDeclinedHandler, this));
            UCConnect.events.on('groupcallstatus', $.proxy(this.groupCallStatusHandler, this));

            Aggregator.on('conversation:delete', $.proxy(this.deleteConversationHandler, this));
            Aggregator.on('conversation:show', $.proxy(this.conversationStartHandler, this));
            Aggregator.on('conversation:state', $.proxy(this.conversationStateHandler, this));

            Aggregator.on('call:init', $.proxy(this.conversationStartHandler, this));
            Aggregator.on('call:accept', $.proxy(this.callAcceptHandler, this));
            Aggregator.on('call:decline', $.proxy(this.callDeclineHandler, this));
            Aggregator.on('call:hangup', $.proxy(this.callHangupHandler, this));

            Aggregator.on('phone:call', $.proxy(this.phoneCallHandler, this));
            Aggregator.on('phone:hangup', $.proxy(this.phoneHangupHandler, this));
            Aggregator.on('phone:hold', $.proxy(this.phoneHoldHandler, this));
            Aggregator.on('phone:dtmf', $.proxy(this.phoneDtmfHandler, this));

            Aggregator.on('groupcall:init', $.proxy(this.groupCallStartHandler, this));
            //Aggregator.on('groupcall:ready', $.proxy(this.groupCallReadyHandler, this));
            //Aggregator.on('groupcall:cancel', $.proxy(this.groupCallCancelHandler, this));
            Aggregator.on('groupcall:accept', $.proxy(this.groupCallAcceptHandler, this));
            Aggregator.on('groupcall:rejoin', $.proxy(this.groupCallAcceptHandler, this));
            Aggregator.on('groupcall:decline', $.proxy(this.groupCallDeclineHandler, this));
            Aggregator.on('groupcall:sendInvite', $.proxy(this.groupCallSendInviteHandler, this));

            Aggregator.on('group:leave', $.proxy(this.groupLeaveHandler, this));
            Aggregator.on('group:join', $.proxy(this.groupJoinHandler, this));
            Aggregator.on('group:create', $.proxy(this.groupCreateHandler, this));
            Aggregator.on('group:invite', $.proxy(this.groupInviteHandler, this));
            Aggregator.on('group:subject', $.proxy(this.groupSubjectHandler, this));

            Aggregator.on('groupconversation:changed', $.proxy(this.groupConversationChangedHandler, this));

            Aggregator.on('session:status', $.proxy(this.sessionStatusHandler, this));
            Aggregator.on('session:setup', $.proxy(this.sessionSetupHandler, this));
            Aggregator.on('session:reloadContacts', $.proxy(this.sessionReloadContactsHandler, this));

            Aggregator.on('roster:add-guest', $.proxy(this.rosterAddGuestHandler, this));
            Aggregator.on('roster:remove-guest', $.proxy(this.rosterRemoveGuestHandler, this));
            Aggregator.on('roster:group:changed', $.proxy(this.rosterGroupChangedHandler, this));
            Aggregator.on('roster:group:create', $.proxy(this.rosterGroupCreateHandler, this));
            Aggregator.on('roster:group:update', $.proxy(this.rosterGroupUpdateHandler, this));
            Aggregator.on('roster:group:delete', $.proxy(this.rosterGroupDeletedHandler, this));
            Aggregator.on('roster:group:toggle', $.proxy(this.rosterGroupToggledHandler, this));

            Aggregator.on('message:send', $.proxy(this.sendMessageHandler, this));

            Aggregator.on('history:refresh', $.proxy(this.refreshCallHistoryHandler, this));

            Aggregator.on('callhistory:received', $.proxy(this.callHistoryReceivedHandler, this));
            Aggregator.on('xmpp-history:loaded', $.proxy(this.fetchOfflineMessages, this));

            // Settings
            Settings.on('change:showChatNotifications', this.handleSettingsChatNotifications);
            Aggregator.on('conversation:showNotifications', $.proxy(this.handleConversationNotifications, this));
            // Aggregator.on('xmpp-history:loaded', $.proxy(this.setupConversationNotificationSettings, this));
            // Aggregator.on('muc-history:loaded', $.proxy(this.setupConversationNotificationSettings, this));
            Aggregator.on('conversation:setupNotifications', $.proxy(this.setupConversationNotificationSettings, this));
            //Aggregator.on('scroll-up', $.proxy(this.testHandler, this));

            // OX
            ox.on('app:init', $.proxy(this.onAppChange, this));
            ox.on('app:resume', $.proxy(this.onAppChange, this));

            Session.setup();

            Roster.attach();

            if (Capabilities.has('messenger-phone')) {
                Phone.attach();
            }

            // Check logged user's contac data update
            ContactsApi.events.on('update', function(event, data) {
                if (Session.user.get('contact').id == data.user_id) {
                    Session.user.setContact(data);
                }
            });

            window.addEventListener('unload', function (event) {
                UCConnect.disconnect();
            });

            $(window).on('roster:groups:reset', $.proxy(this.rosterGroupResetHandler, this));

            this.patchOXUserSettings();

            // required to add AMD module to requirejs
            /*requirejs.config({
             'shim': {
             'com.voiceworks/ox-messenger/core/vendor/promise': []
             }
             });*/
        },

        checkBeforeClosing: function () {
            //this.disconnect();
            //return true;
        },

        /*logHandler: function (level, category, label, content) {
         console.info('proxying to logger', content);
         Session.logger.write(level, category, label, content);
         //this.logger.write(level, category, label, content);
         },*/

        /*
         * Messenger event handlers
         * ================================================
         */
        sessionSetupHandler: function (data) {
            this.log('sessionSetupHandler', data);

            // Set XMPP nickname to allow guest users to see OX user's full name
            UCConnect.setNickname(Session.user.getDisplayName());

            //@TODO: handle result
            var result = UCConnect.connect(data);

            //ApiCall.setParameters(data.apiLocation, data.apiKey, data.sessionId);
            Roster.init(data.userObj);

            if (Capabilities.has('messenger-phone')) {
                Phone.init(data.userObj);
            }
        },

        sessionStatusHandler: function (data) {
            this.log('sessionStatusHandler', data);

            if (data.status.toLowerCase() !== 'offline' && Session.user.get('status').toLowerCase() === 'offline') {
                Aggregator.trigger('connection:reconnecting', {});
                Session.set('xmppHistoryLoaded', false);
                Session.setup();
                return;
            }

            if (this.isOnline()) {

                if (data.status.toLowerCase() !== Session.user.get('status').toLowerCase()) {
                    Session.user.setStatus(data.status, data.statusMessage);
                    UCConnect.status(data.status, data.statusMessage);
                }
            }
        },

        sessionReloadContactsHandler: function () {
            this.log('sessionReloadContactsHandler');

            Session.reloadContacts(Session);
        },

        archiveCollectionHandler: function (event) {
            this.log('archiveCollectionHandler', event);

            var archives = event.data.archive;

            if (event.data.archive.length < 1) {
                Session.set('xmppHistoryLoaded', true);
                Aggregator.trigger('xmpp-history:loaded');
            }

            for (var i = 0; i < archives.length; i++) {
                var bareJid = Util.getBareJid(archives[i].jid);
                var resource = Util.getResourceFromJid(archives[i].jid);

                var conversation = this.startConversation(bareJid, 'chat');
                var archiveModel = new MessageArchiveModel({
                    archive: archives[i],
                    user: Roster.getUserById(bareJid),
                    jid: bareJid,
                    resource: resource,
                    startDate: archives[i].startDate
                });
                conversation.get('jidToDate').push({
                    jid: archives[i].jid,
                    date: archives[i].startDate
                });
                // TODO: should be handled by the setter
                if (!conversation.get('startDate') || archives[i].startDate < conversation.get('startDate')) {
                    conversation.set('startDate', archives[i].startDate);
                }

                conversation.set('timestamp', new Date(archives[i].startDate).getTime());

                conversation.messageArchives.add(archiveModel);
                conversation.messageArchives.sort();
            }

            Session.user.conversations.each(function (conversation) {
                if (!conversation.isMuc()) {
                    conversation.getNextArchiveMessage();
                }
            });

            // TODO execute once all the messageCountCb are finished
            Session.set('xmppHistoryLoaded', true);
            Aggregator.trigger('xmpp-history:loaded');
            this.sortConversations(null);
        },

        callHistoryReceivedHandler: function (data) {
            this.log('callHistoryReceivedHandler', data);

            // Fetch XMPP history for 1:1 conversations
            UCConnect.fetchHistory('xmpp', '', null);

            // Fetch XMPP history for MUCs
            UCConnect.fetchPrivateXml('xmpp', 'muc:rooms', $.proxy(this.mucHistoryHandler, this));
        },

        //messageCollectionHandler: function (data) {
        //    this.log('messageCollectionHandler', data);
        //
        //    var that = this,
        //        conversationWith,
        //        conversation,
        //        remainingMessageCount = 0;
        //
        //    if (!data.length) {
        //        return;
        //    }
        //
        //
        //    _.each(data, function (archive) {
        //        var message,
        //            from = Util.getBareJid(archive.from),
        //            to = Util.getBareJid(archive.to),
        //            historyObject = {
        //                from: from,
        //                username: from,
        //                message: archive.body,
        //                timestamp: archive.timestamp.getTime(),
        //                isHistory: true
        //            };
        //
        //        if (Util.isSameJid(to, UCConnect.getConnection().jid)) {
        //            conversationWith = from;
        //        } else {
        //            conversationWith = to;
        //        }
        //
        //        message = MessageFactory.create(historyObject);
        //
        //        if (!that.msgArchives[conversationWith]) {
        //            that.msgArchives[conversationWith] = [];
        //        }
        //
        //        that.msgArchives[conversationWith].push(message);
        //    });
        //
        //    this.msgArchiveCount[conversationWith]--;
        //
        //    if (this.msgArchiveCount[conversationWith] < 1) {
        //        conversation = this.startConversation(conversationWith, 'chat');
        //
        //        var user = Roster.getUserById(conversationWith),
        //            callHistoryMessages = conversation.createCallHistory(Session.user.callHistory.getContactHistory(user));
        //
        //        // Add Call History
        //        this.msgArchives[conversationWith] = this.msgArchives[conversationWith].concat(callHistoryMessages);
        //
        //        // Sort messages to support conversation 'lastMsg'
        //        this.msgArchives[conversationWith] = _.sortBy(this.msgArchives[conversationWith], function (message) {
        //            return message.get('timestamp');
        //        });
        //
        //        _.each(this.msgArchives[conversationWith], function (message) {
        //            conversation.addMessage(message);
        //        });
        //
        //        this.msgArchives = _.without(this.msgArchives, conversationWith);
        //
        //        for (var i in this.msgArchiveCount) {
        //            if (this.msgArchiveCount[i] > remainingMessageCount) {
        //                remainingMessageCount = this.msgArchiveCount[i];
        //            }
        //        }
        //        if (remainingMessageCount === 0) {
        //            if (!Session.get('xmppHistoryLoaded')) {
        //                Session.set('xmppHistoryLoaded', true);
        //                Aggregator.trigger('xmpp-history:loaded');
        //                this.sortConversations(null);
        //            }
        //        }
        //    }
        //},

        fetchOfflineMessages: function () {
            this.log('fetchOfflineMessages');

            UCConnect.getOfflineMessages('xmpp', true);
        },

        deleteConversationHandler: function () {
            this.log('deleteConversationHandler');

            if (!Conversation.listView.selection) {
                return;
            }

            var selectedItems = Conversation.listView.selection.get(),
                mucChanged = false,
                prevCallConversation,
                currentCallConversation = Session.user.getCurrentCallConversation();

            _.each(selectedItems, function (selectedItem) {
                var selectedConversation = Session.user.conversations.get(selectedItem);

                if (currentCallConversation && selectedConversation.get('id') === currentCallConversation.get('id')) {
                    prevCallConversation = selectedConversation;
                    CoreNotification.yell('warning', gt('The conversation with an active call was not deleted.'));
                    Session.user.conversations.currentConversationId = null;
                    return;
                }

                if (Session.user.conversations.currentConversationId === selectedConversation.id) {
                    Session.user.conversations.currentConversationId = null;
                }

                if (selectedConversation.isMuc()) {
                    UCConnect.groupLeave('xmpp', selectedConversation.get('roomHost'), selectedConversation.get('roomId'));
                    Session.user.conversations.remove(selectedConversation);
                    mucChanged = true;
                } else {
                    Conversation.delete(selectedItem);
                    // TODO remove all the jids present in the conversation
                    var jidToDate = selectedConversation.get('jidToDate');
                    jidToDate.forEach(function (item) {
                        UCConnect.deleteConversation('xmpp', item.jid, item.date);
                    });
                }
            });

            if (mucChanged) {
                Session.user.conversations.trigger('reset');
                Conversation.conversationDeleted();
                Aggregator.trigger('groupconversation:changed', {});
            }

            if (prevCallConversation) {
                Conversation.showConversation(prevCallConversation);
                Conversation.listView.selection.set([prevCallConversation.id]);
            } else {
                var firstConversation = Session.user.conversations.first();
                if (firstConversation) {
                    Conversation.showConversation(firstConversation);
                    Conversation.listView.selection.set([firstConversation.id]);
                }
            }
            Aggregator.trigger('conversationlist:selection:change', []);
        },

        conversationDeletedHandler: function (data) {
            this.log('conversationDeletedHandler', data);

        },

        mucHistoryHandler: function (data, iq) {
            this.log('mucHistoryHandler', data, iq);

            var isIE10 = (navigator.appVersion.indexOf('MSIE 10') !== -1),
                isIE11 = (navigator.appVersion.indexOf('Trident/7') !== -1),
                isEdge = (navigator.appVersion.indexOf('Edge') !== -1),
                mucRooms;

            if (isIE10 || isIE11 || isEdge) {
                if (data && data.childNodes.length) {
                    mucRooms = $.parseJSON(data.childNodes[0].nodeValue);
                }
            } else {
                if (data && data.innerHTML) {
                    mucRooms = $.parseJSON(data.innerHTML);
                }
            }

            _.each(mucRooms, function (room) {
                var conversation = new GroupConversationModel({
                    id: room.roomId + '@' + room.roomHost,
                    roomId: room.roomId,
                    roomName: Util.b64_to_utf8(room.roomName),
                    roomHost: room.roomHost,
                    title: Util.b64_to_utf8(room.roomName),
                    myRoom: room.myRoom,
                    subject: Util.b64_to_utf8(room.subject),
                    user: Session.user,
                    sortName: Util.b64_to_utf8(room.roomName),
                    presenceReceived: false,
                    timestamp: room.timestamp,
                });

                _.each(room.participants, function (participant) {
                    var user = Roster.getUserById(participant);
                    conversation.addParticipant(user, true, true);
                });

                Session.user.conversations.add(conversation);
                if (room.isMember) {
                    Aggregator.trigger('group:join', {
                        conversation: conversation,
                        isHistory: true
                    });
                }
            });
            Session.user.conversations.sort();
            Session.user.conversations.trigger('reset');
            Aggregator.trigger('muchistory:loaded');
        },

        groupConversationChangedHandler: function (data) {
            this.log('groupConversationChangedHandler', data);

            var mucRooms = [];

            _.each(Session.user.conversations.fetchMucRooms(), function (conversation) {
                var participants = conversation.participants.pluck('id'),
                    isMember = conversation.participants.get(Session.user) ? true : false;

                mucRooms.push({
                    roomId: conversation.get('roomId'),
                    roomHost: conversation.get('roomHost'),
                    roomName: Util.b64EncodeUnicode(conversation.get('roomName')),
                    subject: Util.b64EncodeUnicode(conversation.get('subject')),
                    myRoom: conversation.get('myRoom'),
                    participants: participants,
                    isMember: isMember,
                    timestamp: conversation.get('timestamp')
                });
            });
            UCConnect.storePrivateXml('xmpp', 'muc:rooms', '<value>' + JSON.stringify(mucRooms) + '</value>');
            //this.sortConversations();
        },

        memberListHandler: function (event) {
            var conversation = Session.user.conversations.get(event.data.room);
            if (conversation) {
                var previousParticipantList = [],
                    currentParticipantsList = event.data.memberList.map(Util.getBareJid),
                    participantsLeft = [],
                    participantsJoined = [],
                    firstLoad = conversation.participants.length === 0,
                    user;

                conversation.participants.each(function (participant) {
                    previousParticipantList.push(participant.id);
                });
                _.uniq(previousParticipantList);

                participantsLeft = _.difference(previousParticipantList, currentParticipantsList);
                participantsJoined = _.difference(currentParticipantsList, previousParticipantList);

                _.each(participantsLeft, function (participant) {
                    user = Roster.getUserById(participant);
                    conversation.removeParticipant(user);
                    if (Session.user.id === participant) {
                        conversation.properties.set('groupMember', false);
                    }
                });
                _.each(participantsJoined, function (participant) {
                    user = Roster.getUserById(participant);
                    // If it's firstLoad, we don't create system messages
                    conversation.addParticipant(user, false, firstLoad);
                    if (Session.user.id === participant) {
                        conversation.properties.set('groupMember', true);
                    }
                });
                Aggregator.trigger('groupconversation:changed', {});
            }
        },

        refreshCallHistoryHandler: function () {
            this.log('refreshCallHistoryHandler');

            Phone.view.historyView.getCdrs();
        },

        rosterAddGuestHandler: function (data) {
            this.log('rosterAddGuestHandler', data);

            var that = this;

            // Fetch roster with provisioned guest
            UCConnect.getRoster('xmpp');

            Aggregator.on('roster:created', function () {
                // TODO unbind only this listener, no all of them
                Aggregator.off('roster:created');

                var user = Roster.getUserById(data.username);

                // user.properties.addProperty('Guest Users', true);
                user.properties.addProperty(3, true); // 3 = Guest Users

                UCConnect.rosterUserGroupsSet('xmpp', user.get('id'), user.properties.toArray());

                // (Temporarily) store the guest user in the roster with the chosen nickname and e-mail address
                user.set({
                    nickname: _.escape(data.name),
                    contact: {
                        email1: data.email1
                    }
                });
            });
        },

        rosterRemoveGuestHandler: function (data) {
            var that = this,
                question;

            question = gt('Do you really want to delete this guest-user?');

            require([
                'io.ox/core/tk/dialogs',
                'io.ox/core/http',
                'io.ox/contacts/api',
                'io.ox/core/notifications'
            ], function (dialogs, http, ContactsApi, notifications) {
                new dialogs.ModalDialog()
                    .text(question)
                    .addPrimaryButton('delete', gt('Delete'), 'delete', {'tabIndex': '1'})
                    .addButton('cancel', gt('Cancel'), 'cancel', {'tabIndex': '1'})
                    .show()
                    .done(function (action) {
                        if (action === 'delete') {
                            Mediator.removeGuestData(data.user.id);
                            http.GET({
                                module: 'messenger',
                                params: {
                                    action: 'delete_guest_user',
                                    session: ox.session,
                                    mail: data.user.id
                                }
                            }).done(function (result) {
                                ContactsApi.getByEmailaddress(data.user.id).done(function (contact) {
                                    ContactsApi.remove({
                                        folder_id: contact.folder_id,
                                        id: contact.id
                                    }).done(function (contact) { })
                                        .fail(function (e) {
                                            notifications.yell('error', gt('Error deleting guest account contact for %s', data.user.id));
                                            console.error(e);
                                        });
                                });
                            })
                                .fail(function (e) {
                                    notifications.yell('error', gt('Error deleting guest account for %s', data.user.id));
                                });
                        }
                    });
            });
        },

        // handleFetchRosterGroupsToggled: function (data) {
        //     this.log('handleFetchRosterGroupsToggled', data);

        //     if (!data) return;

        //     var groups       = Roster.getGroups(),
        //         storedGroups = JSON.parse(data.textContent);

        //     if (!groups || !storedGroups) return;

        //     groups.each(function (group) {
        //         var newState     = storedGroups[group.id],
        //             currentState = group.get('toggled');

        //         if (newState !== currentState) {
        //             group.set('toggled', newState);
        //         }
        //     });
        // },

        // Triggered when the properties model changes
        // @TODO: loop???
        rosterGroupChangedHandler: function (object) {
            this.log('rosterGroupChangedHandler', object);

            UCConnect.rosterUserGroupsSet('xmpp', object.user.get('id'), object.values);
        },

        handleFetchRosterGroups: function (data) {
            this.log('handleFetchRosterGroups', JSON.parse(data.textContent));

            var storedGroups = JSON.parse(data.textContent);

            // Find missing groups between PrivateXML and Roster
            var rosterGroupIds = Roster.getGroups().models.map(function (group) { return parseInt(group.id, 10); }),
                storedGroupIds = _.keys(storedGroups).map(function (group) { return parseInt(group, 10); }),
                missingGroups  = _.difference(storedGroupIds, rosterGroupIds);

            // Add missing groups
            _.each(missingGroups, function (groupId) {
                var group = Roster.createGroup(storedGroups[groupId]);
            });

            // Set names from PrivateXML for all groups
            var groups = Roster.getGroups();
            groups.each(function(group) {
                var storedGroup = storedGroups[group.id];

                if (storedGroup) {
                    group.set({
                        name    : storedGroup.name,
                        toggled : storedGroup.toggled
                    });
                }
            });

            Session.user.groups.reset(groups.models);

            Roster.groups.sort();
        },

        rosterGroupResetHandler: function () {
            this.log('rosterGroupResetHandler');

            if (_.url.hash('debug-messenger') !== 'true') { return; }

            var jid = Session.user.get('id');
            var groupMapping = {
                1: {
                    toggled : false,
                    name    : gt('Buddylist')
                },
                2: {
                    toggled : false,
                    name    : gt('Favorites')
                },
                3: {
                    toggled : false,
                    name    : gt('Guest Users')
                },
            };

            UCConnect.rosterUserGroupsSet('xmpp', jid, [1,2,3]);
            UCConnect.storePrivateXml('xmpp', 'roster:groups', JSON.stringify(groupMapping));
            UCConnect.getRoster('xmpp');
        },

        /**
         * Handle the creation of a new custom group
         * @param  {Object} object containing the group model and groupName
         * @return {null}
         */
        rosterGroupCreateHandler: function (object) {
            this.log('rosterGroupCreateHandler', object.groupName);

            var newGroup    = Roster.createGroup(object.groupName), // Add the new Group to the Roster
                groupsObj   = Roster.getGroupsObject(), // Get the group Object, IDs and values
                groupKeys   = Roster.getGroupKeys(),
                groupValues = Roster.getGroupValues(),
                users       = Roster.getAllUsers();

            // Save the groupName for this ID in the PrivateXML
            UCConnect.storePrivateXml('xmpp', 'roster:groups', JSON.stringify(groupsObj));

            // Update the groups in the Session User
            Session.user.groups.reset(Roster.getGroups().models);

            if (users) {
                // Add new property for each user, so it's in the dropdown
                users.each(function (user) {
                    user.properties.addProperty(newGroup.get('name'), false);
                });
            }

            Roster.groups.sort();
        },


        /**
         * Handle the update for a new custom group
         * @param  {Object} object containing the group model and new groupName
         * @return {null}
         */
        rosterGroupUpdateHandler: function (object) {
            this.log('rosterGroupUpdateHandler', object);

            var that         = this,
                group        = object.group,
                groupId      = object.group.id,
                groupName    = object.group.get('name'),
                newGroupName = object.groupName,
                newGroup     = Roster.updateGroup(group, newGroupName),
                groupsObj    = Roster.getGroupsObject(),
                groupKeys    = Roster.getGroupKeys(),
                groupValues  = Roster.getGroupValues(),
                users        = Roster.getGroupUsers(groupId);

            // Save the groupName for this ID in the PrivateXML
            UCConnect.storePrivateXml('xmpp', 'roster:groups', JSON.stringify(groupsObj));

            // Update the groups in the Session User
            Session.user.groups.reset(Roster.getGroups().models);

            Roster.groups.sort();
        },

        // Delete the Group Name
        //
        // object.group = Group Model
        rosterGroupDeletedHandler: function (object) {
            this.log('rosterGroupDeletedHandler', object);

            var that         = this,
                group        = object.group,
                groupId      = object.group.id,
                groupName    = object.group.get('name'),
                newGroupName = object.groupName,
                users        = Roster.getGroupUsers(groupId),
                deletedGroup = Roster.deleteGroup(groupId),
                groupsObj    = Roster.getGroupsObject(),
                groupKeys    = Roster.getGroupKeys(),
                groupValues  = Roster.getGroupValues();

            // Save the groupName for this ID in the PrivateXML
            UCConnect.storePrivateXml('xmpp', 'roster:groups', JSON.stringify(groupsObj));

            // Update the groups in the Session User
            Session.user.groups.reset(Roster.getGroups().models);

            if (users) {
                // Add new property for each user, so it's in the dropdown
                users.each(function (user) {
                    // that.deleteGroupForUser(user, groupName);
                    user.properties.deleteProperty(groupId);
                    UCConnect.rosterUserGroupsSet('xmpp', user.get('id'), user.properties.toArray());
                });
            }

            Roster.groups.sort();
        },

        // Save the toggle state of the group
        //
        // object.model = RosterGroupModel
        // object.toggled = Boolean
        rosterGroupToggledHandler: function (object) {
            this.log('rosterGroupToggledHandler', object);

            var groupToggleStatuses,
                group        = Roster.getGroup(object.model.id),
                currentState = object.model.get('toggled'),
                newState     = !currentState;

            group.set('toggled', newState);

            groupToggleStatuses = Roster.getGroupsObject();

            UCConnect.storePrivateXml('xmpp', 'roster:groups', JSON.stringify(groupToggleStatuses));
        },

        removeGuestData: function (user) {
            this.log('removeGuestData', user);

            UCConnect.rosterUserRemove('xmpp', user);
            //todo: decide if we also want to remove the chat history
        },

        conversationStartHandler: function (data) {
            this.log('conversationStartHandler', data);

            var id = data.id,
                type = data.type,
                conversation = Session.user.conversations.get(id),
                self = this;

            if (!conversation) {
                conversation = this.startConversation(id, type);
            } else {
                conversation.setUnreadMessages(0); // fix for not resetting unread messages counter when the conversation already exists.
            }

            Roster.hide();

            require(['com.voiceworks/ox-messenger/main'], function (main) {
                main.getApp().launch().then(function () {
                    main.getApp().showConversation(conversation);
                    // TODO: other side isCallable? or both?
                    if ((type === 'audio' || type === 'video') && Session.user.isCallable()) {
                        Session.user.call.set({
                            callSession: UCConnect.makeCall(id, type === 'video'),
                            type: type,
                            conversationId: id
                        });
                        Session.user.setInCall(true, conversation);
                    }
                    self.sortConversations(id);
                    Aggregator.trigger('conversationlist:selection:change', id);
                });
            });
        },

        conversationStateHandler: function (data) {
            if(Settings.get('showUsersMyChatState')){
                var isGroup = data.conversation instanceof GroupConversationModel;
                var to = data.conversation.id;
                var state = data.state;

                if(!isGroup) {
                    to = to + '/oxmessenger';
                }

                UCConnect.chatState(to, state, isGroup);
            }
        },

        sendMessageHandler: function (message) {
            this.log('sendMessageHandler', message);

            // Send message to UCConnect
            if (message.conversation.isMuc()) {
                UCConnect.groupMessage('xmpp', message.conversation.get('roomHost'), message.conversation.get('roomId'), message.get('message'), message.getFileDescAsObject());
            } else {
                var fullJid = message.conversation.id;
                for (var i in Roster.getUserById(message.conversation.id).get('statuses')) {
                    if (Roster.getUserById(message.conversation.id).get('statuses')[i].status !== 'offline') {
                        fullJid = i;
                        break;
                    }
                }
                UCConnect.message(fullJid, true, message.get('message'), null, message.getFileDescAsObject());
            }
            this.sortConversations();
        },

        groupLeaveHandler: function (data) {
            this.log('groupLeaveHandler', data);

            var conversation = data.conversation;

            conversation.removeParticipant(Session.user);

            conversation.properties.set('groupMember', false);

            return UCConnect.groupLeave('xmpp', conversation.get('roomHost'), conversation.get('roomId'));
        },

        groupJoinHandler: function (data) {
            this.log('groupJoinHandler', data);

            var conversation = data.conversation;

            if (conversation.messageContainers.length) {
                conversation.messageContainers.reset();
            }

            conversation.addParticipant(Session.user, (data.isHistory === undefined) ? false : true);

            conversation.properties.set('groupMember', true);

            UCConnect.groupJoin('xmpp', conversation.get('roomHost'), conversation.get('roomId'), UCConnect.getConnection().jid);
            UCConnect.fetchMemberList('xmpp', conversation.get('roomHost'), conversation.get('roomId'));
            return;
        },

        /*
         * OX event handlers
         * ================================================
         */
        offlineHandler: function (event) {
            this.log('offlineHandler', event);

            var disconnect = this.disconnect.bind(this, event);
            disconnect();
            var timeout = 2000;
            var reconnect = function (event) {
                // TODO this doesnt always reflects the connection status?
                if (Session.user.get('status').toLowerCase() === 'offline') {
                    Aggregator.trigger('connection:reconnecting', event);
                    disconnect();
                    setTimeout(function () {
                        // TODO this can be async-unsafe
                        Session.setup();
                        timeout *= 2;
                        setTimeout(reconnect, timeout);
                    }, 500);
                }
            };
            ox.one('connection:online', function () {
                // wait until network connection will stabilize itself
                // TODO figure out a better way
                setTimeout(reconnect, timeout);
            });
        },

        reloginHandler: function (event) {
            this.log('reloginHandler', event);

            this.disconnect(event);
            ox.on('relogin:success', function (event) {
                Session.setup();
            });
        },


        /*
         * UC event handlers
         * ================================================
         */

        callStateHandler: function (event) {
            this.log('callStateHandler', event, event.data.session.display_name, event.data.status);

            if (Session.user.call.isGroup() && event.data.session.display_name === 'send' && event.data.status === 'answered') {
                // Got an 'answered' status from the send application: group call initiated
                var conversation = Session.user.conversations.get(Session.user.call.get('conversationId')),
                    call = event.data.call,
                    type = call.callType,
                    callType = (type === 'video') ? 'groupvideo' : 'groupaudio',
                    self = this;

                Session.user.call.setCallState('group-initiated');

                this.handleSpeakingEvent(conversation, call);

                if (conversation.get('initiator') !== Session.user) {
                    UCConnect.groupCallReady('xmpp', conversation.get('roomHost'), conversation.get('roomId'), type);

                    conversation.callParticipants.each(function (participant) {
                        if (participant.id === Session.user.id) {
                            return;
                        }

                        if (participant.call.get('readyReceived') === false) {
                            return;
                        }

                        // Call the receive application with the id of the participant
                        if (!participant.call.isActive()) {
                            var recvSession = UCConnect.makeCall('recv', (type === 'video'), ['X-HASH: ' + participant.id + ';' + conversation.get('roomId')], self.groupCallEndCallHandler);
                            participant.call.set('callSession', recvSession);
                        }
                    });

                } else {
                    if (Session.user.call.get('rejoin') === true) {
                        UCConnect.groupCallReady('xmpp', conversation.get('roomHost'), conversation.get('roomId'), type);
                        conversation.callParticipants.each(function (participant) {
                            if (participant.id === Session.user.id) {
                                return;
                            }

                            if (participant.call.get('readyReceived') === false) {
                                return;
                            }

                            // Call the receive application with the id of the participant
                            if (!participant.call.isActive()) {
                                var recvSession = UCConnect.makeCall('recv', (type === 'video'), ['X-HASH: ' + participant.id + ';' + conversation.get('roomId')], self.groupCallEndCallHandler);
                                participant.call.set('callSession', recvSession);
                            }
                        });
                    } else {
                        // Create list of invited users
                        var userlist = [];
                        conversation.participants.each(function (participant) {
                            var error = Session.user.canCallWith(type, participant, true);
                            if (participant !== Session.user && error !== true) {
                                return;
                            }
                            userlist.push({
                                user: participant.id,
                                active: Util.isSameJid(participant.id, UCConnect.getConnection().jid),
                                readyReceived: false
                            });
                        });
                        conversation.getOtherParticipants().each(function (participant) {
                            var error = Session.user.canCallWith(type, participant, true);
                            if (error !== true) {
                                return;
                            }
                            UCConnect.groupCallInvite('xmpp', conversation.get('roomHost'), conversation.get('roomId'), participant.get('id'), type, JSON.stringify(userlist));
                        });
                    }
                }
                this.setGroupCallStatus(conversation, true, type);
                return;
            }

            if (Session.user.call.isGroup() && event.data.session.display_name === 'recv') {
                var conversation = Session.user.conversations.get(Session.user.call.get('conversationId'));
                conversation.callParticipants.each(function (participant) {
                    if (!participant.call.isActive() || participant.call.get('callSession').id !== event.data.session.id) {
                        return;
                    }

                    switch (event.data.status) {
                        case 'trying':
                        case 'in-progress':
                            participant.call.setCallState('in-progress');
                            break;
                        case 'answered':
                            participant.call.setCallState('answered');
                            Util.stopAllSounds();
                            break;
                        case 'ended':
                            participant.call.setCallState('idle');
                            break;
                    }
                });
                return;
            }

            Session.user.call.setCallState(event.data.status);
        },

        handleSpeakingEvent: function (conversation, call) {
            this.log('handleSpeakingEvent', conversation, call);

            var listener = function (data) {
                var roomId = conversation.get('roomId');
                var mucHost = Settings.get('groupConversationRoomHost');

                // TODO handle no call-type scenario from #630
                // https://jira.voiceworks.com/browse/OXM-630
                if (!conversation.get('callType'))
                    return;

                UCConnect.groupCallStatus('xmpp', mucHost, roomId, JSON.stringify({
                    // TODO this is a dirty dirty hack
                    callType: conversation.get('callType').replace('group', ''),
                    speaking: data
                }));
            };

            call.mediaHandler.on('speaking', listener);
            call.once('bye', call.mediaHandler.off.bind(
                call.mediaHandler, 'speaking', listener));
        },

        setGroupCallStatus: function (conversation, hasActiveCall, callType) {
            this.log('setGroupCallStatus', conversation, hasActiveCall, callType);

            assert(conversation instanceof ConversationModel, 'Param type mismatch');

            Session.user.set('inGroupCall', conversation.get('roomId'));

            // if (!callType && hasActiveCall) {
            //     callType = conversation.get('callType');
            // }

            var json = {
                roomId: conversation.get('roomId'),
                status: hasActiveCall
            };

            if (callType) {
                assert(CALL_TYPE[callType]);
                // TODO this is a dirty dirty hack
                callType = callType.replace('group', '');
                json.callType = callType;
            }

            UCConnect.groupCallStatus('xmpp', Settings.get('groupConversationRoomHost'), conversation.get('roomId'),
                JSON.stringify(json));
        },

        connectingHandler: function (event) {
            this.log('connectingHandler', event);

            this.setSessionServerState(event);
            if (event.proto !== 'xmpp') {
                return;
            }

            Session.user.setStatus('offline');
        },

        connectedHandler: function (event) {
            this.log('connectedHandler', event);

            this.setSessionServerState(event);
            if (event.proto !== 'xmpp') {
                return;
            }

            // Remove all conversations
            Session.user.conversations.reset();

            if (Settings.get('initialPresenceStatus')) {
                Session.user.setStatus(Settings.get('initialPresenceStatus'));
            }

            Session.set('connection', 'online');

            CallHistory.fetch(Settings.get('callHistoryLimit'));

            Session.user.set({
                canChat: UCConnect.capabilities.chat,
                canAudio: UCConnect.capabilities.audio,
                canVideo: UCConnect.capabilities.video,
                mustFlash: UCConnect.capabilities.mustFlash
            });
        },

        disconnectedHandler: function (event) {
            this.log('disconnectedHandler', event);

            this.setSessionServerState(event);

            Session.user.setStatus('offline');
            Session.set('connection', 'offline'); // normally a user won't go offline

            this.disconnect();
        },

        endCallHandler: function (event) {
            this.log('endCallHandler', event);

            var conversationId = Session.user.call.get('conversationId') ? Session.user.call.get('conversationId') : Session.user.call.previous('conversationId'),
                conversation = Session.user.conversations.get(conversationId),
                pstnRegExp = new RegExp('@' + Settings.get('pstnCallDomain') + '$', 'i'),
                session = event.data.session,
                isPstn = session && event.data.session.sip_uri.match(pstnRegExp),
                cause = event.data.cause ? event.data.cause : '';

            if (!conversation && event.data.call && event.data.call.request.extraHeaders) {
                conversationId = Util.findRoomIdFromHeaders(event.data.call.request.extraHeaders);
                conversation = Session.user.conversations.get(conversationId + '@' + Settings.get('groupConversationRoomHost'));
            }

            if (!session) {
                throw new Error('SIP session missing');
            }

            if (Session.user.call.isGroup()) {
                var participantsToRemove = [];

                if (session.display_name === 'send') {
                    Session.user.call.end();
                    Session.user.setInCall(false);

                    conversation.callParticipants.each(function (participant) {
                        participant.call.end();
                    });

                    conversation.callParticipants.reset();

                    this.setGroupCallStatus(conversation, false);

                    return;
                }

                if (session.display_name === 'recv') {
                    var participant = conversation.callParticipants.find(function (p) {
                        return p.call.get('callSession') && (p.call.get('callSession').id === event.data.session.id);
                    });

                    if (participant) {
                        participant.call.end();
                        conversation.callParticipants.remove(participant);
                    }
                }

                // Hang up if you are the last one in the group call
                if (conversation.callParticipants.length < 2) {
                    Session.user.call.end();
                    Session.user.setInCall(false);

                    conversation.callParticipants.reset();
                    this.setGroupCallStatus(conversation, false);
                } else {
                    this.setGroupCallStatus(conversation, true, conversation.get('callType'));
                }

                return;
            }

            Session.user.endCall();

            var localHost = event.data.call.localIdentity.uri.host,
                localUser = event.data.call.localIdentity.uri.user,
                remoteHost = event.data.call.remoteIdentity.uri.host,
                remoteUser = event.data.call.remoteIdentity.uri.user,
                isInbound = event.data.call.direction === 'inbound',
                isFailed = event.data.session.status === 'failed',
                historyItem = new HistoryModel({
                    billmsec: (isFailed) ? '0' : event.data.call.endTime.getTime() - event.data.call.startTime.getTime(),
                    call_direction: isInbound ? 'inbound' : 'outbound',
                    effective_destination_domain: isInbound ? localHost : remoteHost,
                    effective_destination_user: isInbound ? localUser : remoteUser,
                    effective_from_domain: isInbound ? remoteHost : localHost,
                    effective_from_user: isInbound ? remoteUser : localUser,
                    hangup_cause: (isFailed) ? 'CALL_REJECTED' : 'NORMAL_CLEARING',
                    start_stamp: (isFailed) ? new Date().toISOString() : new Date(event.data.call.startTime).toISOString()
                });

            // Add call item to Call History and conversation
            Session.user.callHistory.add(historyItem);

            // Do not add Call History items for PSTN calls and group calls
            if (!isPstn && event.data.session.display_name !== 'send' && event.data.session.display_name !== 'recv') {
                if (!conversation) {
                    conversation = this.startConversation(historyItem.getOtherParty(), 'chat');
                    this.sortConversations(null);
                }

                conversation.addMessage(conversation.createCallHistoryItem(historyItem));
            }

            if (conversation && conversation.isMuc()) {
                var frozenParticipants = conversation.participants.filter(function(p) {
                    return p.get('lastCallFrozen') === true;
                });

                if (frozenParticipants.length>0) {
                    CoreNotification.yell('warning', gt('The call was interrupted due to network problems. Please verify that you have a stable internet connection.'));
                }

                // conversation.unset('lastCallFrozen');
            } else if (conversation && !conversation.get('lastCallFrozen')) {
                if (cause !== 'NORMAL_CLEARING') {
                    switch (cause) {
                        case 'RECOVERY_ON_TIMER_EXPIRE':
                            CoreNotification.yell('warning', gt('There seems to be an issue with your network connection which caused the call to end. Please try again.'));
                            break;
                        case '':
                            CoreNotification.yell('warning', gt('There seems to be an issue with the network connection of the other party which caused the call to end. Please try again.'));
                            break;
                    }
                }
            } else if (conversation && conversation.get('lastCallFrozen')) {
                CoreNotification.yell('warning', gt('The call was interrupted due to network problems. Please verify that you have a stable internet connection.'));
                conversation.unset('lastCallFrozen');


            }

            Session.user.setInCall(false);

            //@TODO add event handler for failing PSTN

            if (isFailed) {
                Util.playSound('busy.mp3', 'audioBusy', false);
            }

            if ($('#videoTmp').length) {
                $('#videoTmp').remove();
            }

            if ($('.messenger--notifier-container[data-uri="' + event.data.session.sip_uri + '"]').length) {
                Notification.call.destroy();
            }
        },

        failedHandler: function (event) {
            this.log('failedHandler', event);

            CoreNotification.yell('error', gt('There was a problem connecting Messenger. Please, try again later.'));

            //this.setSessionServerState(event);
            //Session.set('status', 'offline');
        },

        localMediaRequestedHandler: function (event) {
            this.log('localMediaRequestedHandler', event);

            if (!this.mediaRequestTimer && Session.user.call.get('callState') !== 'group-initiated') {
                this.mediaRequestTimer = setTimeout(function () {
                    CoreNotification.yell('info', gt('Please allow access to your microphone and/or webcam'));
                }, 1000);
            }
        },

        localMediaFailedHandler: function (event) {
            this.log('localMediaFailedHandler', event);

            if (this.mediaRequestTimer) {
                window.clearTimeout(this.mediaRequestTimer);
            }
            CoreNotification.yell('error', gt('Access to your microphone and/or webcam failed'));
            delete this.mediaRequestTimer;
        },

        localMediaAcquiredHandler: function (event) {
            this.log('localMediaAcquiredHandler', event);

            if (this.mediaRequestTimer) {
                window.clearTimeout(this.mediaRequestTimer);
            }
            delete this.mediaRequestTimer;
        },

        messageHandler: function (event) {
            this.log('messageHandler()', event, event.data);

            var message = MessageFactory.create(event.data),
                self = this,
                conversation;

            if (message.getSenderJid() == UCConnect.getConnection().jid &&
                message.get('stamp') === '' && !message.get('isCarbonCopy')) {
                return;
            }

            if (message.get('stamp') !== '') {
                message.set('timestamp', new Date().setISO8601(message.get('stamp')).getTime());
                message.set('isHistory', true);
            } else {
                message.set('timestamp', new Date().getTime());
            }

            if (MessageFactory.isSystem(message)) {
                // system messages are silent and modify stuff like room details
                // for now we only deal with room name change coming in with 'subject'
                if (!message.has('subject') || !message.get('subject').length) {
                    return;
                }
                conversation = Session.user.conversations.get(message.getRoomJid());
                conversation.set('subject', message.get('subject'));
                return;
            } else if (MessageFactory.isGroup(message)) {
                conversation = Session.user.conversations.get(message.getRoomJid());
            } else { // chat
                if (message.get('isCarbonCopy')) {
                    conversation = this.startConversation(message.get('carbonTo'), 'chat');
                } else {
                    conversation = this.startConversation(message.getSenderJid(), 'chat');
                }
            }


            // Show desktop and call notifications based on settings
            NotificationControl.notify(conversation, {
                desktop: function () {
                    if (message.get('isHistory') || message.get('message') === '') {
                        return;
                    }

                    self.desktopNotification.show({
                        title: message.user.getDisplayName(),
                        message: message.get('message'),
                        tag: message.getSenderJid(),
                        icon: message.user.getImageUrl(),
                        callback: function () {
                            Aggregator.trigger('conversation:show', { id: (MessageFactory.isGroup(message)) ? message.getRoomJid() : message.getSenderJid(), type: 'chat' });
                            try
                            {
                                window.focus();
                            }
                            catch (ex) { }
                        }
                    });
                }
            });

            conversation.addMessage(message);

            this.sortConversations();
        },

        sortConversations: function (id) {
            this.log('sortConversations', id);

            var tmpSelection;
            if (Conversation.listView) {
                tmpSelection = Conversation.listView.selection.get();
            }

            if (id) {
                tmpSelection = [id];
            }
            Session.user.conversations.sort();
            Session.user.conversations.trigger('reset');
            if (Conversation.listView) {
                Conversation.listView.selection.set(tmpSelection);
            }
        },

        myStatusHandler: function (event) {
            this.log('myStatusHandler', event);
            var resourceId = Util.getResourceFromJid(event.data.username);

            // sync status changes from remote resources (but skip the initial status)
            var localResource = Util.getResourceFromJid(UCConnect.getConnection().jid);
            if (parseInt(event.data.status.priority, 10) > 1 &&
                (event.data.status.status !== Session.user.get('status')) &&
                resourceId != localResource) {
                Session.user.setStatus(event.data.status.status);
                UCConnect.status(event.data.status.status, event.data.status.message);
            }

            // Update user data based on vcard
            if (event.data.vcard) {
                var avatarUrl = event.data.vcard.avatarUrl;

                if (avatarUrl) {
                    Session.user.setImageUrl(avatarUrl);
                    UserPart.updateAvatars(Session.user);
                }
            }

            UserPart.updatePresenceBadges(Session.user.id, event.data.status.status);
        },

        newCallHandler: function (event) {
            this.log('newCallHandler', event);

            var sip_uri = event.data.session.sip_uri,
                user = Roster.getUserById(sip_uri),
                userDisplayName = user.getDisplayName(),
                self = this,
                type = event.data.call.callType, //request.body.match('m=video') !== null ? 'video' : 'audio',
                typetxt = '',
                conversation = Session.user.conversations.get(user.get('jid'));

            assert(CALL_TYPE[type], 'Call type incorrect');

            switch (type) {
                case 'video':
                    typetxt = gt('%s is video calling you...', userDisplayName);
                    break;
                case 'audio':
                    typetxt = gt('%s is calling you...', userDisplayName);
                    break;
                case 'groupvideo':
                    typetxt = gt('%s is group video calling you...', userDisplayName);
                    break;
                case 'groupaudio':
                    typetxt = gt('%s is group calling you...', userDisplayName);
                    break;
            }

            // Show desktop and call notifications based on settings
            NotificationControl.notify(conversation, {
                desktop: function () {
                    self.desktopNotification.show({
                        title: userDisplayName,
                        message: typetxt,
                        tag: sip_uri,
                        icon: user.getImageUrl(),
                        callback: function () {
                            try {
                                window.focus();
                            } catch (ex) { }
                        }
                    });
                },
                call: function () {
                    Notification.call.create(user, type, event.data.session, conversation);
                }
            });

            // @TODO when to ignore this?
            // @TODO rewrite from field in UCConnect Library to show full from instead of only user part of remote identity
        },

        callAcceptHandler: function (event) {
            this.log('callAcceptHandler', event);

            var that = this;

            require(['com.voiceworks/ox-messenger/main'], function (main) {
                main.getApp().launch().then(function () {
                    var conversationId = event.session.sip_uri,
                        conversationType = event.session.call.callType, //request.body.match('m=video') !== null ? 'video' : 'audio',
                        conversation = that.startConversation(conversationId, conversationType);

                    main.getApp().showConversation(conversation);

                    assert(CALL_TYPE[conversationType]);

                    Session.user.call.set({
                        callSession: event.session,
                        type: conversationType,
                        conversationId: conversationId
                    });

                    event.session.answerCall(conversationType === 'video');

                    // Dont start a new call if one is already ongoing.
                    if (!Session.user.isCallable()) {
                        return;
                    }
                    Session.user.setInCall(true, conversation);
                    that.sortConversations(conversation.id);
                });
            });
        },

        callDeclineHandler: function (event) {
            this.log('callDeclineHandler', event);

            //Session.user.call.end();
            try {
                event.session.terminateCall();

            } catch (e) {
                console.error('Call termination error', e);
            }
            //todo: add system message in conversation to log declined call;
        },

        callHangupHandler: function (data) {
            this.log('callHangupHandler', data);

            if (!Session.user.call.isActive()) {
                return;
            }

            if (Session.user.call.isGroup()) {
                var otherUserCnt = 0;
                var participantsToRemove = [];

                // Hangup all participants
                data.conversation.callParticipants.each(function (participant) {
                    if (participant === Session.user) {
                        return;
                    }

                    if (participant.call.isActive()) {
                        participant.call.end();
                        otherUserCnt++;
                    }
                    participantsToRemove.push(participant);
                });

                data.conversation.callParticipants.reset();

                // Remove all invites
                data.conversation.groupCallInvitees.reset();

                // Cancel group call if you are the originator and you are not connected to other participants yet
                if (data.conversation.get('initiator') === Session.user && otherUserCnt === 0) {
                    UCConnect.groupCallCancel('xmpp', data.conversation.get('roomHost'), data.conversation.get('roomId'), Session.user.call.get('type'));
                }
                this.setGroupCallStatus(data.conversation, false);
            }

            Session.user.call.end();
            Session.user.setInCall(false);
        },

        parametersHandler: function (event) {
            this.log('parametersHandler', event);
        },

        initialRosterHandler: function (event) {
            this.log('initialRosterHandler', event);

            Roster.create(event.data.roster);

            UCConnect.events.on('roster', $.proxy(_.debounce(this.rosterHandler, 1000), this));

            UCConnect.fetchPrivateXml('xmpp', 'roster:groups', $.proxy(this.handleFetchRosterGroups, this));
        },

        rosterHandler: function (event) {
            this.log('rosterHandler', event);

            Roster.create(event.data.roster);

            UCConnect.fetchPrivateXml('xmpp', 'roster:groups', $.proxy(this.handleFetchRosterGroups, this));
        },

        setSessionServerState: function (event) {
            this.log('setSessionServerState', event);

            var label = 'server_' + event.proto;
            Session.set(label, event.type);

            if (event.proto !== 'xmpp') {
                return;
            }

            if (event.type === 'disconnected') {
                Aggregator.trigger('session:disconnected', event);
            }
        },

        statusHandler: function (event) {
            this.log('statusHandler', event);

            if (event.data.muc) {
                this.groupStatusHandler(event);
            } else {
                // TODO dirty hack to drop the in-call status for offline users
                // var resources = Object.keys(event.data.status);
                // if (resources.length == 1 && event.data.status[resources[0]].status == 'offline') {
                //     var offlineUser = Roster.getUserById(resources[0]);

                //     offlineUser.set('inGroupCall', false);
                //     Session.user.conversations.each(function (conversation) {
                //         conversation.removeComposingParticipant(offlineUser);
                //     });
                // }
                Roster.setStatus(event.data);
            }
        },

        vcardUpdatedHandler: function (event) {
            this.log('statusHandler', event);
            var user = Roster.getUserById(event.data.username),
                avatarUrl = event.data.vcard.avatarUrl;

            if (avatarUrl) {
                user.setImageUrl(avatarUrl);
                UserPart.updateAvatars(user);
            }
        },

        chatStateHandler: function (event) {
            var conversationJid = Util.getBareJid(event.data.from);
            var isGroup = event.data.isGroup;
            var state = event.data.state;
            var theConversation;
            var fromJid;
            var fromUser;

            // Find the conversation
            Session.user.conversations.each(function (conversation) {
                if(conversation.id === conversationJid) {
                    theConversation = conversation;
                }
            });

            // If the conversation exist
            if(theConversation){
                if(isGroup){
                    fromJid = Util.getResourceFromJid(event.data.from, true);
                }
                else{
                    fromJid = conversationJid;
                }


                // Find the participant
                theConversation.participants.each(function (participant) {
                    if(participant.id === fromJid){
                        fromUser = participant;
                    }
                });

                // If the participant exists (and is NOT me), update conversation composing participants
                if(fromUser && fromUser.id !== Session.user.id){
                    if(state === ConversationModel.STATE_COMPOSING){
                        theConversation.addComposingParticipant(fromUser);
                    }
                    else{
                        theConversation.removeComposingParticipant(fromUser);
                    }
                }
            }
        },

        subscribeHandler: function (event) {
            this.log('subscribeHandler', event);

            if (event.data.username !== '') {
                var pendingInvitations = [].concat(Settings.get('pendingInvitations'));
                pendingInvitations.push({ jid: event.data.username });
                Settings.set('pendingInvitations', pendingInvitations);
                Settings.save();
                Aggregator.trigger('buddylist:reset-invitations');
            }
        },

        groupCreateHandler: function (data) {
            this.log('groupCreateHandler', data);

            var that = this,
                isModerated = false,
                conversation = data.conversation;

            if (!conversation) {
                CoreNotification.yell('error', gt('Group could not be created'));
            }

            var jid = UCConnect.getConnection().jid;
            var success = UCConnect.groupCreate('xmpp', conversation.get('roomHost').toLowerCase(),
                conversation.get('roomId'), conversation.get('roomName'), jid,
                isModerated);

            if (!success) {
                CoreNotification.yell('error', gt('Group could not be created'));
            }

            setTimeout(function () {
                var succesfullyAdded = 0;

                _.each(data.participants, function (participant) {
                    // Do not invite yourself
                    if (Util.isSameJid(participant.id, UCConnect.getConnection().jid)) {
                        return;
                    }

                    var isInvited = that.groupInviteHandler({
                        conversation: conversation,
                        user: participant,
                        notify: false
                    });

                    if (isInvited) {
                        succesfullyAdded++;
                    }
                });

                if (succesfullyAdded === 0 && data.participants.length > 0) {
                    CoreNotification.yell('error', gt('Group could not be created'));
                }

                Aggregator.trigger('groupconversation:changed', {});
            }, 500);
        },

        groupInviteHandler: function (data) {
            this.log('groupInviteHandler', data);

            var conversation = data.conversation,
                user = data.user,
                notify = data.notify,
                success = false,
                callType = Session.user.call.get('type');

            success = UCConnect.groupInvite('xmpp', conversation.get('roomHost').toLowerCase(),
                conversation.get('roomId'), conversation.get('roomName'), user.id);

            if (success) {
                if (notify) {
                    CoreNotification.yell('info', gt('%s has been added to the conversation', user.getDisplayName()));
                    return;
                }
                return true;
            } else {
                CoreNotification.yell('error', gt('Error adding %s', user.getDisplayName()));
            }

            return false;
        },

        groupInvitedHandler: function (event) {
            this.log('groupInvitedHandler', event);

            var data = event.data,
                roomJid = data.roomId + '@' + data.host;

            if (Session.user.conversations.get(roomJid) && Session.user.conversations.get(roomJid).participants.get(Util.getBareJid(UCConnect.getConnection().jid))) {
                return;
            }

            var conversation = new GroupConversationModel({
                id: roomJid,
                roomId: data.roomId,
                roomName: data.roomName,
                roomHost: data.host,
                title: data.roomName,
                myRoom: false,
                user: Session.user,
                timestamp: new Date().getTime(),
                sortName: data.roomName
            });

            Session.user.conversations.add(conversation);

            UCConnect.groupJoin('xmpp', event.data.host, event.data.roomId, UCConnect.getConnection().jid);
            UCConnect.fetchMemberList('xmpp', event.data.host, event.data.roomId);
            this.sortConversations();
        },

        groupStatusHandler: function (event) {
            this.log('groupStatusHandler', event.data.username, event);

            var fixedRooms = [].concat(Settings.get('fixedGroups'));

            var conversation = Session.user.conversations.get(event.data.roomId + '@' + event.data.host),
                user = Roster.getUserById(event.data.username);

            // temp workaround to fix room history size
            if (Util.isSameJid(event.data.username, UCConnect.getConnection().jid) &&
                    event.data.affiliation === 'owner' &&
                    fixedRooms.indexOf(event.data.roomId + '@' + event.data.host) < 0) {
                UCConnect.tmpFixRoomLength('xmpp', event.data.host, event.data.roomId);
                fixedRooms.push(event.data.roomId + '@' + event.data.host);
                Settings.set('fixedGroups', fixedRooms);
            }

            var currentRoomId = Session.user.get('inGroupCall');
            if (!conversation || Util.isSameJid(event.data.username, UCConnect.getConnection().jid) ||
                    !currentRoomId || currentRoomId != conversation.get('roomId')) {
                return;
            }

            conversation.sendGroupCallInvite(user);

            var groupCallStatus = Boolean(conversation.callParticipants &&
                conversation.callParticipants.get(Util.getBareJid(UCConnect.getConnection().jid)));
            var callType = '';

            if (user.call) {
                callType = user.call.get('type');
                if (!callType && conversation.hasActiveCall()) {
                    callType = conversation.get('callType');
                }
            }

            this.setGroupCallStatus(conversation, groupCallStatus, callType);
        },

        groupSubjectHandler: function (data) {
            this.log('groupSubjectHandler', data);
            var conversation = data.conversation;

            UCConnect.groupSubject(
                'xmpp', conversation.get('roomHost').toLowerCase(),
                conversation.get('roomId'),
                conversation.get('subject')
            );
        },

        /*
         * PSTN handlers
         * ================================================
         */

        phoneCallHandler: function (data) {
            this.log('phoneCallHandler', data);

            // Dont start a new call if one is already ongoing.
            if (Session.user.callActive) {
                return;
            }

            var session = UCConnect.makeCall('sip:' + data.number + '@' + Settings.get('pstnCallDomain'), false);
            Session.user.call.set({
                callSession: session,
                type: 'pstn'
            });

            Session.user.setInCall(true);
        },

        phoneHangupHandler: function (data) {
            this.log('phoneHangupHandler', data);

            Session.user.call.end();
            Session.user.setInCall(false);
        },

        phoneHoldHandler: function (data) {
            this.log('phoneHoldHandler', data);

        },

        phoneDtmfHandler: function (data) {
            this.log('phoneDtmfHandler', data);

            switch (data.key) {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    Util.playSound('dtmf/' + data.key + '.mp3', 'dtmf', false);
                    break;
                case '*':
                    Util.playSound('dtmf/asterisk.mp3', 'dtmf', false);
                    break;
                case '#':
                    Util.playSound('dtmf/pound.mp3', 'dtmf', false);
                    break;
            }
            return Session.user.call.isActive() && Session.user.call.get('callSession').sendDTMF(data.key);
        },

        /*
         * Groupcall handlers
         * ================================================
         */

        // This is a method which is required for bypassing the Chrome bug in group calls. It counts the active number of participants to let JsSIP know if the RTCPeerConnection should be closed
        // or that it should put it in some sort of sleep mode until all calls are ended.
        groupCallEndCallHandler: function (data) {
            this.log('groupCallEndCallHandler', data);

            var xHeader = null,
                conversationId = null,
                conversation = null;

            if (data.remote_identity.uri.user === 'send') {
                return true;
            }

            for (var i in data.request.extraHeaders) {
                if (data.request.extraHeaders[i].indexOf('X-HASH') > -1) {
                    xHeader = data.request.extraHeaders[i];
                    break;
                }
            }
            if (xHeader) {
                xHeader = xHeader.split(';');
                conversationId = xHeader[1];
                conversation = Session.user.conversations.get(conversationId + '@' + Settings.get('groupConversationRoomHost'));
                if (!conversation) {
                    return true; // no conversation found, must be something wrong but the call has to be ended
                }
                if (conversation.callParticipants.length - 1 < 2) {
                    return true; // this was the final receiver session, end call
                } else {
                    return false; // there are more receiver sessions active, do not terminate the call
                }
            } else {
                return true; // no conversation found, must be something wrong but the call has to be ended
            }
            return false;
        },

        groupCallStartHandler: function (data) {
            this.log('groupCallStartHandler', data);

            // Dont start a new call if one is already ongoing.
            if (!Session.user.isCallable()) {
                return;
            }

            var callType = data.type === 'video' ? 'groupvideo' : 'groupaudio',
                warnings = [];

            var sendSession = UCConnect.makeCall('send', data.type === 'video',
                ['X-HASH: ' + Util.getBareJid(Session.user.id) + ';' + data.conversation.get('roomId')],
                this.groupCallEndCallHandler);


            data.conversation.callParticipants.reset();

            // Add yourself to the list of (active) call participants
            data.conversation.addCallParticipant(Session.user, sendSession, true, callType);

            data.conversation.set('initiator', Session.user);

            Session.user.call.set({
                callSession: sendSession,
                type: callType,
                conversationId: data.conversation.id
            });

            data.conversation.getOtherParticipants().each(function (participant) {
                var error = Session.user.canCallWith(data.type, participant, true);
                if (error !== true) {
                    warnings.push(error);
                    return;
                }

                data.conversation.addCallParticipant(participant.id, null, false, callType);
            });

            if (warnings.length) {
                CoreNotification.yell('warning', warnings.join('\n\n'));
            }

            Session.user.setInCall(true, data.conversation);

            // Set a timeout in case all participants ignore the call
            data.conversation.startActivityTimeout(function (activity) {
                if (!activity){
                    Aggregator.trigger('call:hangup', {conversation: data.conversation});
                }
            });
        },

        groupCallSendInviteHandler: function (data) {
            this.log('groupCallSendInviteHandler', data);

            var userlist = [],
                user     = data.user,
                callType = Session.user.call.get('type').replace('group', '');

            var canCallWith = Session.user.canCallWith(callType, user, true);
            if (canCallWith !== true) {
                CoreNotification.yell('info', gt('%s has been added to the conversation, but cannot join the call. He has been asked to join using Chrome or Firefox.', user.getDisplayName()));
                return false;
            }

            data.conversation.callParticipants.each(function (participant) {
                userlist.push({
                    user: participant.id,
                    active: Util.isSameJid(participant.id, UCConnect.getConnection().jid),
                    readyReceived: participant.call.get('readyReceived')
                });
            });

            assert(CALL_TYPE[callType]);
            UCConnect.groupCallInvite('xmpp', data.conversation.get('roomHost'), data.conversation.get('roomId'), data.user.get('id'), callType, JSON.stringify(userlist));
        },

        groupCallInviteHandler: function (event) {
            this.log('groupCallInviteHandler', event);

            // Dont accepts a new call if one is already ongoing.
            if (!Session.user.isCallable()) {
                return;
            }

            Session.user.callRinging = true;

            assert(CALL_TYPE[event.data.type]);

            var callType = event.data.type === 'video' ? 'groupvideo' : 'groupaudio',
                initiator = Roster.getUserById(event.data.username),
                conversation = Session.user.conversations.get(event.data.roomId + '@' + event.data.host);

            if (!conversation) {
                return;
            }

            conversation.callParticipants.reset();

            _.each(event.data.participantlist, function (participant) {
                if (Util.isSameJid(participant.user, UCConnect.getConnection().jid)) {
                    return;
                }

                // Add user to participants list (if necessary)
                if (!conversation.participants.get(participant.user)) {
                    conversation.participants.add(Roster.getUserById(participant.user));
                }

                // Add user to call participants list (if necessary)
                if (!conversation.callParticipants.get(participant.user)) {
                    conversation.addCallParticipant(participant.user, null, participant.user === event.data.username || participant.readyReceived, callType);
                }
            });

            conversation.set('initiator', initiator);

            var tmpSession = {
                sip_uri: conversation.get('id')
            };

            var sip_uri = conversation.get('id'),
                user = Roster.getUserById(sip_uri),
                userDisplayName = initiator.getDisplayName(),
                typetxt = '';

            switch (callType) {
                case 'groupvideo':
                    typetxt = gt('%s is group video calling you...', userDisplayName);
                    break;
                case 'groupaudio':
                    typetxt = gt('%s is group calling you...', userDisplayName);
                    break;
            }

            // Dispose the state on accept or reject, unbind the listener as well.
            var dropRingingState = function () {
                // console.log('dropRingingState');
                Session.user.callRinging = false;
                Aggregator.off('groupcall:accept', dropRingingState);
                Aggregator.off('groupcall:decline', dropRingingState);
                UCConnect.events.removeListener('groupcallcancelled', dropRingingState);
            };
            Aggregator.once('groupcall:accept', dropRingingState);
            Aggregator.once('groupcall:decline', dropRingingState);
            // TODO should be once, but doesnt work
            UCConnect.events.addListener('groupcallcancelled', dropRingingState);

            this.desktopNotification.show({
                title: userDisplayName,
                message: typetxt,
                tag: sip_uri,
                icon: user.getImageUrl(),
                callback: function () {
                    try {
                        window.focus();
                    } catch (ex) { }
                }
            });

            Notification.call.create(initiator, callType, tmpSession, conversation);

            this.setGroupCallStatus(conversation, true, callType);
        },

        groupCallReadyHandler: function (event) {
            this.log('groupCallReadyHandler', event);

            assert(CALL_TYPE[event.data.type]);

            var callType = event.data.type === 'video' ? 'groupvideo' : 'groupaudio',
                conversation = Session.user.conversations.get(event.data.roomId + '@' + event.data.host),
                participant = conversation.callParticipants.get(Util.getBareJid(event.data.from));

            // Ignore self
            if (Util.isSameJid(event.data.from, UCConnect.getConnection().jid)) {
                return;
            }

            if (!participant) {
                var jid = Util.getBareJid(event.data.from);
                conversation.addCallParticipant(jid, null, true, callType);
                participant = conversation.callParticipants.get(jid);
            }

            participant.call.set('readyReceived', true);

            // if you're not a callParticipant, you haven't accepted the invitation (yet)
            if (conversation.callParticipants.get(Util.getBareJid(UCConnect.getConnection().jid))) {
                Util.stopAllSounds();
                var recvSession = UCConnect.makeCall('recv', event.data.type === 'video', ['X-HASH: ' + Util.getBareJid(event.data.from) + ';' + event.data.roomId], this.groupCallEndCallHandler);
                participant.call.set('callSession', recvSession);
            }

            this.setGroupCallStatus(conversation, Session.user.call.isActive(), callType);
        },

        groupCallCancelledHandler: function (event) {
            this.log('groupCallCancelledHandler', event);

            var conversation = Session.user.conversations.get(event.data.username);

            conversation.callParticipants.each(function (participant) {
                participant.call.end();
            });

            Session.user.call.end();
            Session.user.setInCall(false);

            // Remove notification
            if ($('.messenger--notifier-container[data-uri="' + conversation.id + '"]').length) {
                Notification.call.destroy();
            }
            this.setGroupCallStatus(conversation, false);
        },

        groupCallAcceptHandler: function (call) {
            this.log('groupCallAcceptHandler', call);

            assert(CALL_TYPE[call.type]);

            var isVideo = call.type === 'video',
                callType = isVideo ? 'groupvideo' : 'groupaudio',
                self = this;

            var sendSession = UCConnect.makeCall('send', isVideo,
                ['X-HASH: ' + Util.getBareJid(UCConnect.getConnection().jid) + ';' + call.conversation.get('roomId')],
                this.groupCallEndCallHandler);

            Session.user.call.set({
                callSession: sendSession,
                type: callType,
                conversationId: call.conversation.id
            });

            Session.user.setInCall(true, call.conversation);

            call.conversation.addCallParticipant(Session.user, sendSession, true, callType);

            // When Rejoining a Call
            if (call.rejoin) {
                Session.user.call.set('rejoin', true);
                this.groupCallRejoin(call, callType);
            }

            Aggregator.trigger('conversation:show', {
                id: call.conversation.id,
                type: isVideo ? 'groupvideo' : 'groupaudio'
            });

            this.setGroupCallStatus(call.conversation, true, callType);
        },

        groupCallRejoin: function (call, callType) {
            this.log('groupCallRejoin', call, callType);

            assert(CALL_TYPE[callType]);

            var callParticipants = call.conversation.getCallParticipants();

            _.each(callParticipants, function (participant) {
                if (participant === Session.user) {
                    return;
                }
                call.conversation.addCallParticipant(participant.get('jid'), null, true, callType);
                participant.call.set('readyReceived', true);
            });

            if (call.conversation.get('initiator') === Session.user) {
                //UCConnect.groupCallReady('xmpp', call.conversation.get('roomHost'), call.conversation.get('roomId'), call.type);
            }
        },

        groupCallDeclineHandler: function (data) {
            this.log('groupCallDeclineHandler', data);

            assert(CALL_TYPE[data.type]);

            var result = UCConnect.groupCallDecline('xmpp', data.conversation.get('roomHost'), data.conversation.get('roomId'), data.type);
            this.setGroupCallStatus(data.conversation, false);
        },

        groupCallDeclinedHandler: function (event) {
            this.log('groupCallDeclinedHandler', event);

            var conversation = Session.user.conversations.get(event.data.roomId + '@' + event.data.host),
                participant = conversation.callParticipants.get(Util.getBareJid(event.data.from));

            if (!conversation) {
                return;
            }

            if (participant) {
                conversation.callParticipants.remove(participant);
            }

            this.setGroupCallStatus(conversation, conversation.getCallParticipantsCount() > 1,
                conversation.get('callType'));

            // End call if you are the only one left
            if (conversation.getCallParticipantsCount() === 1){
                Aggregator.trigger('call:hangup', {conversation: conversation});
            }
        },

        groupCallStatusHandler: function (event) {
            this.log('groupCallStatusHandler', event);

            var searchJid = event.data.roomId + '@' + event.data.host,
                conversation = Session.user.conversations.get(searchJid);

            if (!conversation) {
                return;
            }

            var participant = conversation.participants.get(Util.getBareJid(event.data.from));
            if (!participant) {
                return;
            }

            var json = JSON.parse(Strophe.xmlunescape(event.data.status));

            if (json.speaking !== undefined) {
                if (Session.user.call.get('callState') === 'idle' && participant.get('isSpeaking') === true) {
                    participant.set('isSpeaking', false);
                } else {
                    participant.set('isSpeaking', !!json.speaking);
                }
            }

            if (json.status !== undefined && (!participant.get('inGroupCall') ||
                    participant.get('inGroupCall') == json.roomId)) {
                var participantsInGroupCall;

                participant.set('inGroupCall', json.status ? json.roomId : '');
                participantsInGroupCall = conversation.getNumParticipantsInGroupCall();

                // Hangup the call if I'm the only one in the group and people are hanging up (n_ready > n_participants)
                // This is a hack for firefox to close the group calls
                // when the other participants hangup and there's no focus in firefox
                // or the permissions window is not accepted
                if (Session.user.id !== participant.id && // Make sure the event is sent by another user
                        Session.user.call.get('callState') === 'group-initiated' &&
                        participantsInGroupCall === 1 &&
                        conversation.getNumParticipantsReady() > participantsInGroupCall) {
                    this.log('groupCall:hangup', 'Only one participant remaining');
                    Aggregator.trigger('call:hangup', {conversation: conversation});
                }

            }

            if (json.callType) {
                assert(CALL_TYPE[json.callType]);
                conversation.set('callType', json.callType);
            }
        },

        /*
         * Helpers
         * ================================================
         */
        inviteUser: function (jid) {
            this.log('inviteUser', jid);

            if (!UCConnect.rosterUserAdd('xmpp', jid)) {
                return false;
            }
            if (!UCConnect.rosterUserSubscribe('xmpp', jid)) {
                return false;
            }
            return true;
        },

        authorizeUser: function (jid) {
            this.log('authorizeUser', jid);

            return UCConnect.rosterUserAuthorize('xmpp', jid);
        },

        /**
         *
         * @param email {string}
         * @returns {Promise}
         */
        addContact: function (email) {
            this.log('addContact', email);

            // Adding that buddy as contact in your address book, but be avare that untill he accept the invitation,
            // he will not be on lists for adding in conversations.
            return ContactsApi.create({
                folder_id: CoreSettings.get('folder/contacts'),
                email1: email
            }).done(function (contact) {
                Aggregator.trigger('session:reloadContacts');
            }).fail(function (e) {
                console.error(e);
            });
        },

        unauthorizeUser: function (jid) {
            this.log('unauthorizeUser', jid);

            return UCConnect.rosterUserUnauthorize('xmpp', jid);
        },

        disconnect: function (event) {
            this.log('disconnect', event);

            Session.set('connection', 'offline');
            UCConnect.SessionHandler.terminateAll();

            Aggregator.trigger('session:disconnected', event);

            UCConnect.disconnect();
        },

        isOnline: function () {
            this.log('isOnline');

            return Session.get('connection') === 'online';
        },

        startConversation: function (id, type) {
            this.log('startConversation', id, type);

            var participant = Roster.getUserById(id);
            return Conversation.start(participant, type);
        },

        handleSettingsChatNotifications: function (event) {
            this.log('handleSettingsChatNotifications', event);

        },

        handleConversationNotifications: function (event) {
            this.log('handleConversationNotifications', event);

            var savedSettings = Settings.get('conversationNotificationSettings'),
                newSettings = [],
                conversation = event.conversation,
                conversationJid = conversation.get('id');

            // Remove all matches for this jid
            var newSettings = _.reject(savedSettings, function (s) {
                return s.jid === conversationJid;
            });

            // Save setting
            newSettings.push({
                'jid': conversationJid,
                'showChatNotifications': event.showChatNotifications
            });

            Settings.set('conversationNotificationSettings', newSettings);
            Settings.save();
        },

        setupConversationNotificationSettings: function (event) {
            this.log('setupConversationNotificationSettings', event);

            if (!event) {
                return;
            }

            var conversationId = event.get('id'),
                conversation = Session.user.conversations.get(conversationId);

            if (!conversation) {
                return;
            }

            var currentSetting = _.findWhere(Settings.get('conversationNotificationSettings'), {jid: conversationId});

            if (currentSetting) {
                conversation.set('showNotifications', currentSetting.showChatNotifications);
                conversation.properties.set('showNotifications', currentSetting.showChatNotifications);
            }
        },

        patchOXUserSettings: function () {
            require(['io.ox/core/settings/user'], function (userSettings) {
                var oldFunc = userSettings.editCurrentUser;

                // Override OX core method
                userSettings.editCurrentUser = function ($node) {
                    var user = oldFunc.call(userSettings, $node);

                    user.then(function (user) {
                        // Listen for the user sync event (triggered when the data is saved on the server)
                        user.once('sync', function (value) {
                            var imageUrl = value.get('image1_url');

                            // The first time we update the image, image1_url is empty!
                            if (!imageUrl) {
                                setTimeout(function () {
                                    imageUrl = '/ajax/image/user/picture?id=' + value.id;
                                    UCConnect.vcard(imageUrl);
                                }, 2000);
                            } else {
                                UCConnect.vcard(imageUrl);
                            }
                        });
                    });

                    return user;
                };
            });
        },

        onAppChange: function (app) {
            // Check if it's a real app change
            if (this._currentAppName !== app.get('name')) {
                // If switch to a different app (not Messenger), keep streams alive
                if ((app.get('name') !== 'com.voiceworks/ox-messenger')){
                    Conversation.keepCurrentAlive();
                }
                // If we resume Messenger, refresh the current conversation
                else if (app.get('state') === 'running') {
                    var currentCallConversation =
                            Session.user.getCurrentCallConversation(),
                        currentConversation =
                            Session.user.conversations.get(Session.user.conversations.currentConversationId);

                    // If active call is the focused one, show it
                    if (currentCallConversation === currentConversation){
                        Conversation.showConversation(currentConversation, false, true);
                    }
                }

                this._currentAppName = app.get('name');
            }
        },

        log: function () {
            if (_.url.hash('debug-messenger') && window.console) {
                console.info('mediator: ' + _.first(arguments), _.rest(arguments));
            }
        }
    };

    return Mediator;
});
