/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2015 Voiceworks B.V. info@voiceworks.com
 *
 */

define('com.voiceworks/ox-messenger/util', [
    'io.ox/core/moment',
    'gettext!com.voiceworks/ox-messenger'
], function (moment, gt) {
    'use strict';

    var Util = {
        getDateFormatted: function (timestamp, options) {

            if (!_.isNumber(timestamp)) {
                return gt('unknown');
            }

            var opt = $.extend({ fulldate: false, filtertoday: true }, options || {}),
                d = moment(timestamp),

                timestr = function () {
                    return d.format('LT');
                },
                datestr = function () {
                    return d.format('l') + (opt.fulldate ? ' ' + timestr() : '');
                },
                isSameDay = function () {
                    return moment().isSame(d, 'day');
                };

            if (opt.filtertoday && isSameDay()) {
                return timestr();
            }

            if (opt.smart) {
                var delta = moment().startOf('day').diff(moment(timestamp).startOf('day'), 'day');
                if (delta === 1) { return gt('Yesterday'); } else if (delta <= 6) { return d.format('dddd'); }
            }

            return datestr();
        },

        getDurationFormatted: function (seconds) {
            if (!_.isNumber(seconds)) {
                return 0;
            }

            var duration = parseInt(seconds, 10),
                hours = Math.floor(duration / 3600),
                minutes = Math.floor((duration - (hours * 3600)) / 60),
                seconds = duration - (hours * 3600) - (minutes * 60);

            if (hours < 10) {
                hours = '0' + hours;
            }
            if (minutes < 10) {
                minutes = '0' + minutes;
            }
            if (seconds < 10) {
                seconds = '0' + seconds;
            }

            if (duration >= 3600) {
                return hours + ':' + minutes + ':' + seconds;
            } else {
                return minutes + ':' + seconds;
            }
        },

        getDateYmdHms: function (date) {
            return date.toISOString().substring(0, 19).replace('T', ' ');
        },

        generateUUID: function () {
            var d = new Date().getTime();
            var uuid = 'xxxxxxxxxxxx4xxxyxxxxxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
                var r = (d + Math.random() * 16) % 16 | 0;
                d = Math.floor(d / 16);
                return (c == 'x' ? r : (r & 0x7 | 0x8)).toString(16);
            });
            return uuid;
        },
        appendQuery: function (url, query) {
            var queryStr = typeof query === 'string' ? query : $.param(query);
            return url + (url.indexOf('?') === -1 ? '?' : '&') + queryStr;
        },
        playSound: function (soundFile, id, loop, maxDuration, isRingtone) {
            var type,
                tmpObject = isRingtone === true ? 'tmpringaudio' : 'tmpaudio';

            this.stopAllSounds(true);

            if (!soundFile) {
                return;
            }

            switch (soundFile.split('.').pop()) {
            case 'wav':
                type = 'audio/wav';
                break;
            case 'mp3':
                type = 'audio/mpeg';
                break;
            default:
            //console.error('audio mime type not supported');
            }

            var range = '';
            if (maxDuration) {
                range = '#t=,' + maxDuration;
            }

            window[tmpObject] = $('<audio>').prop('autoplay', true).attr('id', id);
            if (loop) {
                window[tmpObject].prop('loop', true);
            }
            window[tmpObject].append(
                $('<source>').attr({ src: ox.base + '/apps/com.voiceworks/ox-messenger/sounds/' + soundFile + range, type: type })
                );
        },
        stopAllSounds: function (ringtoneExcluded) {
            if (window.tmpaudio !== undefined) {
                window.tmpaudio.get(0).pause();
                delete window.tmpaudio;
            }

            if (!ringtoneExcluded && window.tmpringaudio !== undefined) {
                window.tmpringaudio.get(0).pause();
                delete window.tmpringaudio;
            }
        },
        phoneNumberFilter: function (number, numbersOnly) {
            if (numbersOnly) {
                // Filter all characters except 0-9
                return number = number.replace('+', '00').replace(/[^0-9]/gim, '');
            }

            // Filter all characters except 0-9, *, + and #
            return number.replace(/[^0-9\*\+#]/gim, '');
        },
        escapeRegExp: function (str) {
            return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, '\\$&');
        },
        getBareJid: function (jid) {
            if (!jid) {
                return '';
            }

            if (jid.indexOf('/') > -1) {
                jid = jid.substring(0, jid.indexOf('/'));
            }

            return jid;
        },
        isSameJid: function (jid1, jid2) {
            return (Util.getBareJid(jid1) == Util.getBareJid(jid2));
        },
        /**
         *
         * @param {string} jid
         * @return {string} Resource starts with '/' or it's an empty string.
         */
        getResourceFromJid: function (jid, filterAll) {
            if (jid.indexOf('/') > -1) {
                jid = jid.substring(jid.indexOf('/'));
                if(filterAll){
                  jid = jid.substr(1);
                  if(jid.indexOf('#') > -1){
                      jid = jid.substring(0, jid.indexOf('#'));
                  }
                }
                return jid;
            }

            return '';
        },

        stringTruncate: function (string, maxLength, suffix) {
            if (string.length > maxLength) {
                string = string.substring(0, maxLength - 1);

                if (string.indexOf(' ') !== -1) {
                    string = string.substring(0, Math.min(string.length - 2, string.lastIndexOf(' ')));
                }

                string = string + suffix;
            }

            return string;
        },
        keyify: function (string) {
            if (!string || typeof string !== 'string') {
                return string;
            }

            return string.toLowerCase().replace(/\W/g, '_');
        },
        byteCount: function(s) {
            return encodeURI(s).split(/%(?:u[0-9A-F]{2})?[0-9A-F]{2}|./).length - 1;
        },
        findRoomIdFromHeaders: function (headers) {
            var matchedHeader = _.find(headers, function (header) {
                return header.indexOf('X-HASH') >= 0;
            });

            if (matchedHeader) {
                return _.last(matchedHeader.split(';'));
            } else {
                return false;
            }
        },


    };

    if (!Date.setISO8601) {
        // from http://stackoverflow.com/questions/5249216/javascript-timestamp-from-iso8061
        Date.prototype.setISO8601 = function (dString) {
            var regexp = /(\d\d\d\d)(-)?(\d\d)(-)?(\d\d)(T)?(\d\d)(:)?(\d\d)(:)?(\d\d)(\.\d+)?(Z|([+-])(\d\d)(:)?(\d\d))/;
            if (dString.toString().match(new RegExp(regexp))) {
                var d = dString.match(new RegExp(regexp));
                var offset = 0;
                this.setUTCDate(1);
                this.setUTCFullYear(parseInt(d[1], 10));
                this.setUTCMonth(parseInt(d[3], 10) - 1);
                this.setUTCDate(parseInt(d[5], 10));
                this.setUTCHours(parseInt(d[7], 10));
                this.setUTCMinutes(parseInt(d[9], 10));
                this.setUTCSeconds(parseInt(d[11], 10));
                if (d[12]) {
                    this.setUTCMilliseconds(parseFloat(d[12]) * 1000);
                }
                else {
                    this.setUTCMilliseconds(0);
                }
                if (d[13] != 'Z') {
                    offset = (d[15] * 60) + parseInt(d[17], 10);
                    offset *= ((d[14] == '-') ? -1 : 1);
                    this.setTime(this.getTime() - offset * 60 * 1000);
                }
            }
            else {
                this.setTime(Date.parse(dString));
            }
            return this;
        };
    }

    return {
        formatTime: function (timestamp) {
            return Util.getDateFormatted(timestamp, { fulldate: false });
        },
        formatDateTime: function (timestamp) {
            return Util.getDateFormatted(timestamp);
        },
        formatFileSize: function (fileSize) {
            var kb = Math.pow(2, 10);
            var mb = Math.pow(2, 20);

            if(fileSize > mb)
                return (fileSize/mb).toFixed(2) + 'Mb';
            else
                return (fileSize/kb).toFixed(0) + 'kb';
        },
        getDurationFormatted: function (seconds) {
            return Util.getDurationFormatted(seconds);
        },
        generateUUID: function () {
            return Util.generateUUID();
        },
        playSound: Util.playSound,
        stopAllSounds: Util.stopAllSounds,
        appendQuery: Util.appendQuery,
        phoneNumberFilter: Util.phoneNumberFilter,
        escapeRegExp: Util.escapeRegExp,
        imageOptions: 'width=128&height=128&scaleType=cover',
        imageOptionsSmall: 'width=80&height=80&scaleType=cover',
        contactPhoneFields: ['cellular_telephone1', 'cellular_telephone2', 'telephone_business1', 'telephone_business2', 'telephone_home1', 'telephone_home2'],
        contactPhoneFieldMapping: { // Copied from io.ox/mail/write/view-main.js
            telephone_business1: gt('Phone (business)'),
            telephone_business2: gt('Phone (business)'),
            telephone_home1: gt('Phone (private)'),
            telephone_home2: gt('Phone (private)'),
            cellular_telephone1: gt('Mobile'),
            cellular_telephone2: gt('Mobile')
        },
        getBareJid: Util.getBareJid,
        getResourceFromJid: Util.getResourceFromJid,
        isSameJid: Util.isSameJid,
        getDateYmdHms: Util.getDateYmdHms,
        b64EncodeUnicode: function (str) {
            return btoa(encodeURIComponent(str).replace(/%([0-9A-F]{2})/g, function (match, p1) {
                return String.fromCharCode('0x' + p1);
            }));
        },
        b64_to_utf8: function (str) {
            return decodeURIComponent(escape(window.atob(str)));
        },
        stringTruncate: Util.stringTruncate,
        keyify: Util.keyify,
        byteCount: Util.byteCount,
        findRoomIdFromHeaders: Util.findRoomIdFromHeaders
    };
});
